//
// Copyright (c) 2008 smdn <smdn@smdn.jp>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Formats.Mime;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    internal Stream OpenReadWithAttributes(ImapFetchDataItem fetchDataItem,
                                           bool setBusy,
                                           out IImapMessageAttribute messageAttributes)
    {
      messageAttributes = null;

      CheckMessageNumberAvailable(SequenceOrUidSet);

      if (setBusy)
        Client.SetBusy();

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      true,
                                                      null,
                                                      null,
                                                      setBusy),
                           fetchDataItem,
                           null,
                           out messageAttributes);
    }

    private static void ValidateAndTranslateFetchOptions(ImapMessagePartInfo messagePart,
                                                         ImapPartialRange? fetchRange,
                                                         ImapMessageFetchBodyOptions options,
                                                         EncodingSelectionCallback selectFallbackCharset,
                                                         out bool fetchPeek,
                                                         out string fetchSection,
                                                         out string encoding,
                                                         out Encoding charset)
    {
      encoding = null;
      charset = null;

      fetchPeek = ((int)(options & ImapMessageFetchBodyOptions.SetSeen) == 0);

      if (messagePart == null || messagePart.Specifier.Length == 0) {
        if ((int)(options & ImapMessageFetchBodyOptions.OmitHeader) == 0)
          fetchSection = null; // fetch entire message
        else
          fetchSection = "TEXT"; // fetch body of message
      }
      else {
        fetchSection = messagePart.Specifier; // fetch specific message part
      }

      if ((int)(options & ImapMessageFetchBodyOptions.DecodeContent) == 0) {
        if (messagePart == null)
          charset = null; // charset is specified by caller
        else
          charset = iso8859_1;
      }
      else { // if decode content
        if (messagePart == null)
          throw new NotSupportedException("ImapMessageInfo + ImapMessageFetchBodyOptions.DecodeContent is not supported");
        if (fetchRange.HasValue)
          throw new NotSupportedException("ImapPartialRange + ImapMessageFetchBodyOptions.DecodeContent is not supported");

        encoding = messagePart.ContentTransferEncoding;

        var charsetString = messagePart.ContentTypeCharset;

        if (charsetString == null)
          charset = iso8859_1;
        else
          charset = EncodingUtils.GetEncodingThrowException(charsetString, selectFallbackCharset);
      }
    }

    private TResult FetchMessageBody<TResult>(ImapMessagePartInfo messagePart,
                                              ImapPartialRange? fetchRange,
                                              ImapMessageFetchBodyOptions options,
                                              EncodingSelectionCallback selectFallbackCharset,
                                              Func<Stream, Encoding, TResult> converter)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      ValidateAndTranslateFetchOptions(messagePart,
                                       fetchRange,
                                       options,
                                       selectFallbackCharset,
                                       out var fetchPeek,
                                       out var fetchSection,
                                       out var encoding,
                                       out var charset);

      Client.SetBusy();

      var stream = PrepareStream(new FetchMessageBodyStream(this,
                                                            fetchPeek,
                                                            fetchSection,
                                                            fetchRange,
                                                            true),
                                 null,
                                 encoding,
                                 out _);

      return converter(stream, charset);
    }

    private Stream FetchMessageBody(ImapMessageFetchBodyOptions options,
                                    CancellationToken cancellationToken,
                                    IProgress<ImapMessageFetchProgressInfo> progress)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      ValidateAndTranslateFetchOptions(null,
                                       null,
                                       options,
                                       null,
                                       out var fetchPeek,
                                       out _,
                                       out _,
                                       out _);

      Client.SetBusy();

      return PrepareStream(new FetchMessageBodyStream(this,
                                                      fetchPeek,
                                                      null,
                                                      null,
                                                      cancellationToken,
                                                      progress),
                           null,
                           null,
                           out _);
    }

    private Task<TResult> FetchMessageBodyAsync<TResult>(ImapMessagePartInfo messagePart,
                                                         // ImapPartialRange? fetchRange,
                                                         ImapMessageFetchBodyOptions options,
                                                         EncodingSelectionCallback selectFallbackCharset,
                                                         CancellationToken cancellationToken,
                                                         IProgress<ImapMessageFetchProgressInfo> progress,
                                                         Func<Stream, Encoding, TResult> converter)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      bool fetchPeek;
      string fetchSection;
      string encoding;
      Encoding charset;

      ValidateAndTranslateFetchOptions(messagePart,
                                       null, // fetchRange
                                       options,
                                       selectFallbackCharset,
                                       out fetchPeek,
                                       out fetchSection,
                                       out encoding,
                                       out charset);

      return Client.StartTaskExclusively(() => {
        var stream = PrepareStream(new FetchMessageBodyStream(this,
                                                              fetchPeek,
                                                              fetchSection,
                                                              null,
                                                              cancellationToken,
                                                              progress),
                                   null,
                                   encoding,
                                   out _);

        if (cancellationToken.IsCancellationRequested)
          stream.Dispose();

        cancellationToken.ThrowIfCancellationRequested();

        return converter(stream, charset);
      }, cancellationToken, false);
    }

    private Stream PrepareStream(FetchMessageBodyStream stream,
                                 ImapFetchDataItem fetchDataItem,
                                 string encoding,
                                 out IImapMessageAttribute attributes)
    {
      try {
        Mailbox.ProcessResult(stream.Prepare(fetchDataItem, out attributes));

        if (encoding == null)
          return stream;
        else
          return ContentTransferEncoding.CreateDecodingStream(stream, encoding);
      }
      catch {
        stream.Dispose();

        throw;
      }
    }

    private TResult FetchHeaderFields<TResult>(ImapMessagePartInfo messagePart,
                                               /*params*/ IEnumerable<string> fetchHeaderFields,
                                               Converter<Stream, TResult> streamConverter)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      Client.SetBusy();

      return FetchHeaderFieldsCore(messagePart,
                                   fetchHeaderFields,
                                   streamConverter);
    }

    private Task<TResult> FetchHeaderFieldsAsync<TResult>(ImapMessagePartInfo messagePart,
                                                          /*params*/ IEnumerable<string> fetchHeaderFields,
                                                          Converter<Stream, TResult> streamConverter)
    {
      CheckMessageNumberAvailable(SequenceOrUidSet);

      return Client.StartTaskExclusively(() => FetchHeaderFieldsCore(messagePart,
                                                                     fetchHeaderFields,
                                                                     streamConverter),
                                         CancellationToken.None, false);
    }

    private TResult FetchHeaderFieldsCore<TResult>(ImapMessagePartInfo messagePart,
                                                   /*params*/ IEnumerable<string> fetchHeaderFields,
                                                   Converter<Stream, TResult> streamConverter)
    {
      var headerFields = string.Join(" ", fetchHeaderFields ?? Enumerable.Empty<string>());

      string fetchSectionAndHeaderFields = null;

      if (messagePart == null || messagePart.Specifier.Length == 0) {
        if (headerFields.Length == 0)
          fetchSectionAndHeaderFields = "HEADER";
        else
          fetchSectionAndHeaderFields = "HEADER.FIELDS (" + headerFields + ")";
      }
      else {
        if (headerFields.Length == 0)
          fetchSectionAndHeaderFields = messagePart.Specifier + ".MIME";
        else
          throw new NotSupportedException();
      }

      using (var stream = new FetchMessageBodyStream(this, fetchSectionAndHeaderFields)) {
        Mailbox.ProcessResult(stream.Prepare(null, out _));

        return streamConverter(stream);
      }
    }

    private sealed class FetchMessageBodyStream : ImapFetchMessageBodyStream {
      public ImapMessageInfo Message {
        get; private set;
      }

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    bool fetchPeek,
                                    string fetchSection,
                                    ImapPartialRange? fetchRange,
                                    bool unsetBusy)
        : base(message.Client.Session,
               fetchPeek,
               message.SequenceOrUidSet,
               fetchSection,
               fetchRange,
               DefaultFetchBlockSize)
      {
        this.Message = message;
        this.unsetBusy = unsetBusy;
      }

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    bool fetchPeek,
                                    string fetchSection,
                                    ImapPartialRange? fetchRange,
                                    CancellationToken cancellationToken,
                                    IProgress<ImapMessageFetchProgressInfo> progress)
        : base(message.Client.Session,
               fetchPeek,
               message.SequenceOrUidSet,
               fetchSection,
               fetchRange,
               DefaultFetchBlockSize,
               cancellationToken)
      {
        this.Message = message;
        this.unsetBusy = true;
        this.progress = progress;
      }

      public FetchMessageBodyStream(ImapMessageInfo message,
                                    string fetchSectionAndHeaderFields)
        : base(message.Client.Session,
               true,
               message.SequenceOrUidSet,
               fetchSectionAndHeaderFields,
               null,
               DefaultFetchBlockSize)
      {
        this.Message = message;
        this.unsetBusy = true;
      }

      protected override void Dispose(bool disposing)
      {
        if (Message != null) {
          if (unsetBusy)
            Message.Client.UnsetBusy();

          Message = null;
        }

        base.Dispose(disposing);
      }

      internal protected override ImapCommandResult Prepare(ImapFetchDataItem fetchDataItem,
                                                            out IImapMessageAttribute messageAttr)
      {
        CheckDisposed();

        try {
          return base.Prepare(fetchDataItem, out messageAttr);
        }
        catch {
          Dispose();

          throw;
        }
      }

      internal protected override void ReportFetchedOctetCount(long fetchedOctetCount, long totalOctetCount)
      {
        if (progress != null)
          progress.Report(new ImapMessageFetchProgressInfo(fetchedOctetCount, totalOctetCount, CancellationToken));
      }

      protected override Exception GetNoSuchMessageException(ImapCommandResult result)
      {
        CheckDisposed();

        Message.Mailbox.ProcessResult(result);

        Message.SetDeleted();

        return new ImapMessageVanishedException(Message);
      }

      protected override Exception GetFetchFailureException(ImapCommandResult result)
      {
        CheckDisposed();

        Message.Mailbox.ProcessResult(result);

        return base.GetFetchFailureException(result);
      }

      private void CheckDisposed()
      {
        if (Message == null)
          throw new ObjectDisposedException(GetType().FullName);
      }

      private readonly bool unsetBusy = true;
      private readonly IProgress<ImapMessageFetchProgressInfo> progress;
    }
  }
}

