// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NETFRAMEWORK || NETSTANDARD2_0
#define SERIALIZATION
#endif

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
#if SERIALIZATION
using System.Runtime.Serialization;
#endif

using Smdn.IO;
using Smdn.IO.Streams;
using Smdn.Text;

namespace Smdn.Net.Imap4.Protocol {
  // handles '4. Data Formats'
#if SERIALIZATION
  [Serializable]
#endif
  public abstract class ImapData
#if SERIALIZATION
    : ISerializable
#endif
  {
    public static ImapData CreateTextData(ByteString text)
    {
      if (text == null)
        throw new ArgumentNullException(nameof(text));

      // 4.1. Atom, 4.2. Number, 4.3. String
      return new ImapTextData(text);
    }

    public static ImapData CreateTextData(Stream textStream)
    {
      if (textStream == null)
        throw new ArgumentNullException(nameof(textStream));

      // 4.1. Atom, 4.2. Number, 4.3. String
      return new ImapTextStreamData(textStream);
    }

    public static ImapData CreateListData(ImapData[] list)
    {
      if (list == null)
        throw new ArgumentNullException(nameof(list));

      // 4.4. Parenthesized List
      return new ImapListData(list);
    }

    private static readonly ImapData nil = new ImapNilData();

    public static ImapData CreateNilData()
    {
      // 4.5. NIL
      return nil;
    }

    /*
     * instance members
     */
    public abstract ImapDataFormat Format { get; }

    public virtual ImapData[] List {
      get { throw CreateNotListException(); }
    }

#if SERIALIZATION
    public virtual void GetObjectData(SerializationInfo info, StreamingContext context)
    {
    }
#endif

    public virtual string GetTextAsString()
    {
      throw CreateNotTextException();
    }

    public virtual ByteString GetTextAsByteString()
    {
      throw CreateNotTextException();
    }

    [CLSCompliant(false)]
    public virtual ulong GetTextAsNumber()
    {
      throw CreateNotTextException();
    }

    public virtual byte[] GetTextAsByteArray()
    {
      throw CreateNotTextException();
    }

    public virtual Stream GetTextAsStream()
    {
      throw CreateNotTextException();
    }

    public virtual long GetTextLength()
    {
      throw CreateNotTextException();
    }

    public virtual void CopyText(byte[] buffer, int offset, int count)
    {
      throw CreateNotTextException();
    }

    public virtual int GetText(ref byte[] buffer)
    {
      throw CreateNotTextException();
    }

    private Exception CreateNotListException()
    {
      return new InvalidOperationException("data is not list");
    }

    private Exception CreateNotTextException()
    {
      return new InvalidOperationException("data is not text");
    }

    /*
     * data classes
     */
#if SERIALIZATION
    [Serializable]
#endif
    private class ImapNilData : ImapData {
      public override ImapDataFormat Format {
        get { return ImapDataFormat.Nil; }
      }

      public ImapNilData()
      {
      }

#if SERIALIZATION
      private ImapNilData(SerializationInfo info, StreamingContext context)
      {
      }
#endif

      public override string ToString()
      {
        return "{NIL}";
      }
    }

#if SERIALIZATION
    [Serializable]
#endif
    private class ImapListData : ImapData {
      public override ImapDataFormat Format {
        get { return ImapDataFormat.List; }
      }

      public override ImapData[] List {
        get { return list; }
      }

      public ImapListData(ImapData[] list)
      {
        this.list = list;
      }

#if SERIALIZATION
      private ImapListData(SerializationInfo info, StreamingContext context)
      {
        this.list = (ImapData[])info.GetValue("list", typeof(ImapData[]));
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue("list", list);
      }
#endif

      public override string ToString()
      {
        return string.Format("{{List:{0}}}", string.Join(", ", list.Select(d => d.ToString())));
      }

      private readonly ImapData[] list;
    }

#if SERIALIZATION
    [Serializable]
#endif
    private class ImapTextData : ImapData {
      public override ImapDataFormat Format {
        get { return ImapDataFormat.Text; }
      }

      public ImapTextData(ByteString text)
      {
        this.text = text;
      }

#if SERIALIZATION
      private ImapTextData(SerializationInfo info, StreamingContext context)
      {
        var t = (byte[])info.GetValue("text", typeof(byte[]));

        this.text = ByteString.CreateImmutable(t);
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue("text", GetTextAsByteArray());
      }
#endif

      public override string GetTextAsString()
      {
        return text.ToString();
      }

      public override ByteString GetTextAsByteString()
      {
        return text;
      }

      public override ulong GetTextAsNumber()
      {
        try {
          return text.ToUInt64();
        }
        catch (FormatException) {
          throw new ImapMalformedDataException(this);
        }
      }

      public override byte[] GetTextAsByteArray()
      {
        return text.ToArray();
      }

      public override Stream GetTextAsStream()
      {
        return new MemoryStream(text.Segment.Array,
                                text.Segment.Offset,
                                text.Segment.Count,
                                false);
      }

      public override long GetTextLength()
      {
        return text.Length;
      }

      public override void CopyText(byte[] buffer, int offset, int count)
      {
        Buffer.BlockCopy(text.Segment.Array, text.Segment.Offset,
                         buffer, offset, count);
      }

      public override int GetText(ref byte[] buffer)
      {
        if (buffer == null || buffer.Length < text.Length)
          buffer = new byte[text.Length];

        Buffer.BlockCopy(text.Segment.Array, text.Segment.Offset,
                         buffer, 0, text.Length);

        return text.Length;
      }

      public override string ToString()
      {
        return string.Format("{{Text:{0}}}", text);
      }

      private readonly ByteString text;
    }

#if SERIALIZATION
    [Serializable]
#endif
    private class ImapTextStreamData : ImapData {
      public override ImapDataFormat Format {
        get { return ImapDataFormat.Text; }
      }

      public ImapTextStreamData(Stream textStream)
      {
        this.textStream = textStream;
      }

#if SERIALIZATION
      private ImapTextStreamData(SerializationInfo info, StreamingContext context)
      {
        var t = (byte[])info.GetValue("text", typeof(byte[]));

        this.textStream = new MemoryStream(t, false); // XXX: this would be allocated on LOH
      }

      public override void GetObjectData(SerializationInfo info, StreamingContext context)
      {
        info.AddValue("text", GetTextAsByteArray()); // XXX: this would be allocated on LOH
      }
#endif

      public override string GetTextAsString()
      {
        return ByteString.CreateImmutable(GetTextAsByteArray()).ToString();
      }

      public override ByteString GetTextAsByteString()
      {
        return ByteString.CreateImmutable(GetTextAsByteArray());
      }

      public override ulong GetTextAsNumber()
      {
        throw new NotImplementedException();
      }

      public override byte[] GetTextAsByteArray()
      {
        try {
          textStream.Position = 0L;

          var chunkedStream = textStream as ChunkedMemoryStream;

          if (chunkedStream == null)
            return textStream.ReadToEnd();
          else
            return chunkedStream.ToArray();
        }
        finally {
          textStream.Position = 0L;
        }
      }

      public override Stream GetTextAsStream()
      {
        return textStream;
      }

      public override long GetTextLength()
      {
        return textStream.Length;
      }

      public override void CopyText(byte[] buffer, int offset, int count)
      {
        try {
          textStream.Position = 0L;
          textStream.Read(buffer, offset, count);
        }
        finally {
          textStream.Position = 0L;
        }
      }

      public override int GetText(ref byte[] buffer)
      {
        if (buffer == null || buffer.Length < textStream.Length)
          buffer = new byte[(int)textStream.Length];

        try {
          textStream.Position = 0L;

          return textStream.Read(buffer, 0, buffer.Length);
        }
        finally {
          textStream.Position = 0L;
        }
      }

      public override string ToString()
      {
        return string.Format("{{Text:<Stream ({0} octets)>}}", textStream.Length);
      }

      private readonly Stream textStream;
    }
  }
}