// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

namespace Smdn.Net.Imap4 {
  // combinable data item types:
  //   ImapCombinableDataItem
  //     => handles 'data item names' etc.
  //     ImapFetchDataItem
  //       => handles 'message data item names or macro'
  //     ImapSearchCriteria
  //       => handles 'searching criteria'
  //     ImapSortCriteria
  //       => handles 'sort criteria'
  //     ImapStatusDataItem
  //       => handles 'status data item names'
  //   * ImapStoreDataItem
  //       => handles 'message data item name' and 'value for message data item'

  // 6.4.6. STORE Command
  //   message data item name
  //   value for message data item
  public sealed class ImapStoreDataItem : ImapCombinableDataItem {
    private static readonly ImapString flagsReplace       = new ImapString("FLAGS");
    private static readonly ImapString flagsReplaceSilent = new ImapString("FLAGS.SILENT");
    private static readonly ImapString flagsAdd           = new ImapString("+FLAGS");
    private static readonly ImapString flagsAddSilent     = new ImapString("+FLAGS.SILENT");
    private static readonly ImapString flagsRemove        = new ImapString("-FLAGS");
    private static readonly ImapString flagsRemoveSilent  = new ImapString("-FLAGS.SILENT");

    private static readonly string[] emptyKeywords =
#if NET45
      ArrayExtensions.Empty<string>();
#else
      Array.Empty<string>();
#endif
    private static readonly ImapMessageFlag[] emptyFlags =
#if NET45
      ArrayExtensions.Empty<ImapMessageFlag>();
#else
      Array.Empty<ImapMessageFlag>();
#endif

    // FLAGS <flag list>
    //     Replace the flags for the message (other than \Recent) with the
    //     argument.  The new value of the flags is returned as if a FETCH
    //     of those flags was done.
    public static ImapStoreDataItem ReplaceFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      return FlagsCommon(flagsReplace, emptyKeywords, flags.Prepend(flag));
    }

    public static ImapStoreDataItem ReplaceFlags(string keyword, params string[] keywords)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));

      return FlagsCommon(flagsReplace, keywords.Prepend(keyword), emptyFlags);
    }

    public static ImapStoreDataItem ReplaceFlags(string[] keywords, ImapMessageFlag[] flags)
    {
      return FlagsCommon(flagsReplace, keywords, flags);
    }

    public static ImapStoreDataItem ReplaceFlags(IImapMessageFlagSet flags)
    {
      return FlagsCommon(flagsReplace, flags);
    }

    // FLAGS.SILENT <flag list>
    //     Equivalent to FLAGS, but without returning a new value.
    public static ImapStoreDataItem ReplaceFlagsSilent(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      return FlagsCommon(flagsReplaceSilent, emptyKeywords, flags.Prepend(flag));
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(string keyword, params string[] keywords)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));

      return FlagsCommon(flagsReplaceSilent, keywords.Prepend(keyword), emptyFlags);
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(string[] keywords, ImapMessageFlag[] flags)
    {
      return FlagsCommon(flagsReplaceSilent, keywords, flags);
    }

    public static ImapStoreDataItem ReplaceFlagsSilent(IImapMessageFlagSet flags)
    {
      return FlagsCommon(flagsReplaceSilent, flags);
    }

    // +FLAGS <flag list>
    //     Add the argument to the flags for the message.  The new value
    //     of the flags is returned as if a FETCH of those flags was done.
    public static ImapStoreDataItem AddFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      return FlagsCommon(flagsAdd, emptyKeywords, flags.Prepend(flag));
    }

    public static ImapStoreDataItem AddFlags(string keyword, params string[] keywords)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));

      return FlagsCommon(flagsAdd, keywords.Prepend(keyword), emptyFlags);
    }

    public static ImapStoreDataItem AddFlags(string[] keywords, ImapMessageFlag[] flags)
    {
      return FlagsCommon(flagsAdd, keywords, flags);
    }

    public static ImapStoreDataItem AddFlags(IImapMessageFlagSet flags)
    {
      return FlagsCommon(flagsAdd, flags);
    }

    // +FLAGS.SILENT <flag list>
    //     Equivalent to +FLAGS, but without returning a new value.
    public static ImapStoreDataItem AddFlagsSilent(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      return FlagsCommon(flagsAddSilent, emptyKeywords, flags.Prepend(flag));
    }

    public static ImapStoreDataItem AddFlagsSilent(string keyword, params string[] keywords)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));

      return FlagsCommon(flagsAddSilent, keywords.Prepend(keyword), emptyFlags);
    }

    public static ImapStoreDataItem AddFlagsSilent(string[] keywords, ImapMessageFlag[] flags)
    {
      return FlagsCommon(flagsAddSilent, keywords, flags);
    }

    public static ImapStoreDataItem AddFlagsSilent(IImapMessageFlagSet flags)
    {
      return FlagsCommon(flagsAddSilent, flags);
    }

    // -FLAGS <flag list>
    //     Remove the argument from the flags for the message.  The new
    //     value of the flags is returned as if a FETCH of those flags was
    //     done.
    public static ImapStoreDataItem RemoveFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      return FlagsCommon(flagsRemove, emptyKeywords, flags.Prepend(flag));
    }

    public static ImapStoreDataItem RemoveFlags(string keyword, params string[] keywords)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));

      return FlagsCommon(flagsRemove, keywords.Prepend(keyword), emptyFlags);
    }

    public static ImapStoreDataItem RemoveFlags(string[] keywords, ImapMessageFlag[] flags)
    {
      return FlagsCommon(flagsRemove, keywords, flags);
    }

    public static ImapStoreDataItem RemoveFlags(IImapMessageFlagSet flags)
    {
      return FlagsCommon(flagsRemove, flags);
    }

    // -FLAGS.SILENT <flag list>
    //     Equivalent to -FLAGS, but without returning a new value.
    public static ImapStoreDataItem RemoveFlagsSilent(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      return FlagsCommon(flagsRemoveSilent, emptyKeywords, flags.Prepend(flag));
    }

    public static ImapStoreDataItem RemoveFlagsSilent(string keyword, params string[] keywords)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));

      return FlagsCommon(flagsRemoveSilent, keywords.Prepend(keyword), emptyFlags);
    }

    public static ImapStoreDataItem RemoveFlagsSilent(string[] keywords, ImapMessageFlag[] flags)
    {
      return FlagsCommon(flagsRemoveSilent, keywords, flags);
    }

    public static ImapStoreDataItem RemoveFlagsSilent(IImapMessageFlagSet flags)
    {
      return FlagsCommon(flagsRemoveSilent, flags);
    }

    // common construction method
    private static ImapStoreDataItem FlagsCommon(ImapString itemName, IImapMessageFlagSet flags)
    {
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      flags = flags.GetNonApplicableFlagsRemoved();

      if (flags.Count == 0)
        throw new ArgumentException("must contain at least one keyword or one applicable flag", nameof(flags));

      return new ImapStoreDataItem(itemName, flags.GetNonApplicableFlagsRemoved().ToArray());
    }

    private static ImapStoreDataItem FlagsCommon(ImapString itemName, string[] keywords, ImapMessageFlag[] flags)
    {
      if (keywords == null)
        throw new ArgumentNullException(nameof(keywords));
      if (flags == null)
        throw new ArgumentNullException(nameof(flags));

      var items = new List<ImapString>(keywords.Length + flags.Length);

      foreach (var flag in flags) {
        if (flag == null)
          throw new ArgumentException("can not contain null", nameof(flags));

        if (!ImapMessageFlag.NonApplicableFlags.Contains(flag))
          items.Add(flag);
      }

      foreach (var keyword in keywords) {
        if (keyword == null)
          throw new ArgumentException("can not contain null", nameof(keywords));

        items.Add(new ImapMessageFlag(keyword, null));
      }

      if (items.Count == 0)
        throw new ArgumentException("must contain at least one keyword or one applicable flag");

      return new ImapStoreDataItem(itemName, items.ToArray());
    }

    public ImapString ItemName {
      get { return itemName; }
    }

    private ImapStoreDataItem(ImapString itemName, ImapString[] items)
      : base(items)
    {
      this.itemName = itemName;
    }

    protected override ImapStringList GetCombined()
    {
      return ToParenthesizedString();
    }

    public override string ToString()
    {
      return string.Concat(itemName, " ", GetCombined());
    }

    private ImapString itemName;
  }
}
