using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionTests {
    [Test]
    public void TestReadOnlyProperty_ChangeCollection_ServerCapabilities()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(ImapCapability.Imap4Rev1));
      });
    }

    [Test]
    public void TestReceiveTimeout_SetZero()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.ReceiveTimeout = 0;

        Assert.AreEqual(0, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestReceiveTimeout_SetInfinite()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.ReceiveTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSendTimeout_SetZero()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.SendTimeout = 0;

        Assert.AreEqual(0, session.SendTimeout);
      });
    }

    [Test]
    public void TestSendTimeout_SetInfinite()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.SendTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.SendTimeout);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetZero()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.TransactionTimeout = 0;

        Assert.AreEqual(0, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetInfinite()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.TransactionTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.TransactionTimeout);
      });
    }

    private void Idle(Action<ImapSession> preprocessSession, Action<ImapSession> processSession)
    {
      TestCases.WithAuthenticatedSession(new[] {"IDLE"}, (session, server) => {
        preprocessSession(session);

        server.EnqueueResponse("IDLE\r\n",
                               "+ idling\r\n");

        using (var cts = new CancellationTokenSource()) {
          var t = session.IdleAsync(cts.Token);

          Assert.IsTrue(session.IsIdling);
          Assert.IsTrue(session.IsTransactionProceeding);

          processSession(session);

          server.EnqueueTaggedResponse("DONE\r\n",
                                       "$tag OK done\r\n");

          cts.Cancel();

          Assert.IsTrue((bool)t.Result);

          Assert.IsFalse(session.IsIdling);
          Assert.IsFalse(session.IsTransactionProceeding);
        }
      });
    }

    [Test]
    public void TestTransactionTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.TransactionTimeout = 1000;
        Assert.AreEqual(1000, session.TransactionTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.TransactionTimeout = 1000);
      });
    }

    [Test]
    public void TestSendTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.SendTimeout = 1000;
        Assert.AreEqual(1000, session.SendTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.SendTimeout = 1000);
      });
    }

    [Test]
    public void TestReceiveTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.ReceiveTimeout = 1000;
        Assert.AreEqual(1000, session.ReceiveTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.ReceiveTimeout = 1000);
      });
    }

    [Test]
    public void TestBeginNewTransaction_WhileTransactionProceeding_1()
    {
      Idle(delegate(ImapSession session) {
        Assert.IsFalse(session.IsTransactionProceeding);
      },
      delegate(ImapSession session) {
        Assert.IsTrue(session.IsTransactionProceeding);
        Assert.Throws<InvalidOperationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestBeginNewTransaction_WhileTransactionProceeding_2()
    {
      TestCases.WithConnectedSession((session, server) => {
        using (var transactionStartedEvent = new ManualResetEvent(false)) {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag ", // incomplete response
                                       null,
                                       () => transactionStartedEvent.Set());

          var noopTask = Task.Run(() => session.NoOp());

          transactionStartedEvent.WaitOne();

          Assert.IsTrue(session.IsTransactionProceeding);

          Assert.Throws<InvalidOperationException>(() => session.NoOp());

          server.EnqueuePushResponse("OK \r\n");

          Assert.DoesNotThrow(() => {
            if (!noopTask.Wait(TimeSpan.FromSeconds(1)))
              Assert.Fail("transaction not stopped");
          });
        }
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.HandlesIncapableAsException = false;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Binary));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapString[] {
          new ImapLiteralString("literal1", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
          new ImapLiteralString("literal2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
          new ImapLiteralString("literal3", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
          new ImapLiteralString("literal4", Encoding.ASCII, ImapLiteralOptions.Literal),
          new ImapLiteralString("literal5", Encoding.ASCII, ImapLiteralOptions.Literal8),
          new ImapLiteralString("literal6", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
          new ImapLiteralString("literal7", Encoding.ASCII, ImapLiteralOptions.Synchronizing | ImapLiteralOptions.Literal8IfCapable),
          new ImapLiteralString("literal8", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable | ImapLiteralOptions.Literal8),
          new ImapParenthesizedString(new ImapLiteralString("nested1", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
                                      new ImapLiteralString("nested2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
                                      new ImapLiteralString("nested3", Encoding.ASCII, ImapLiteralOptions.Literal8),
                                      new ImapLiteralString("nested4", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable)),
        };

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST " +
                        "{8}\r\nliteral1 " +
                        "{8+}\r\nliteral2 " +
                        "{8}\r\nliteral3 " +
                        "{8}\r\nliteral4 " +
                        "~{8}\r\nliteral5 " +
                        "{8}\r\nliteral6 " +
                        "{8}\r\nliteral7 " +
                        "~{8}\r\nliteral8 " +
                        "({7+}\r\nnested1 " +
                        "{7}\r\nnested2 " +
                        "~{7}\r\nnested3 " +
                        "{7}\r\nnested4)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_LiteralNonSyncCapable()
    {
      TestCases.WithConnectedSession(new[] {"LITERAL+"}, (session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapParenthesizedString(new ImapString[] {
          new ImapLiteralString("literal1", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
          new ImapLiteralString("literal2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
          new ImapLiteralString("literal3", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
        });

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST (" +
                        "{8}\r\nliteral1 " +
                        "{8+}\r\nliteral2 " +
                        "{8+}\r\nliteral3)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_LiteralNonSyncIncapable()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));

        Assert.Throws<ImapIncapableException>(delegate {
          session.GenericCommand("X-TEST", new ImapLiteralString("literal", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing));
        });
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_Literal8Capable()
    {
      TestCases.WithConnectedSession(new[] {"BINARY"}, (session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Binary));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapParenthesizedString(new ImapString[] {
          new ImapLiteralString("literal1", Encoding.ASCII, ImapLiteralOptions.Literal),
          new ImapLiteralString("literal2", Encoding.ASCII, ImapLiteralOptions.Literal8),
          new ImapLiteralString("literal3", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
        });

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST (" +
                        "{8}\r\nliteral1 " +
                        "~{8}\r\nliteral2 " +
                        "~{8}\r\nliteral3)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_Literal8Incapable()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Binary));

        Assert.Throws<ImapIncapableException>(delegate {
          session.GenericCommand("X-TEST", new ImapLiteralString("literal", Encoding.ASCII, ImapLiteralOptions.Literal8));
        });
      });
    }

    [Test]
    public void TestConnect()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.IsFalse(session.IsIdling);
        Assert.IsNull(session.SelectedMailbox);
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(string.Format("imap://{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
      });
    }

    [Test]
    public void TestConnect_Bye()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* BYE not available.\r\n",
                                   null, null, true);

        var ex = Assert.Throws<ImapConnectionException>(() => {
          using (var session = new ImapSession(server.Host, server.Port)) {
          }
        });

        Assert.IsNull(ex.InnerException);
        StringAssert.Contains("not available.", ex.Message);
      });
    }

    [Test]
    public void TestConnect_InvalidResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("+OK pseudo POP server ready.\r\n");

        var ex = Assert.Throws<ImapConnectionException>(() => {
          using (var session = new ImapSession(server.Host, server.Port)) {
          }
        });

        Assert.IsNull(ex.InnerException);
        StringAssert.Contains("unexpected data response", ex.Message);
      });
    }

    [Test]
    public void TestConnect_TransactionTimeout()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse(null); // no greeting response

        var ex = Assert.Throws<ImapConnectionException>(() => {
          using (var session = new ImapSession(server.Host, server.Port, 500)) {
          }
        });

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      });
    }

    [Test]
    public void TestConnect_ConnectionTimeout()
    {
      var endPoint = new IPEndPoint(IPAddress.Loopback, 0);

      using (var server = new Socket(endPoint.AddressFamily, SocketType.Stream, ProtocolType.Tcp)) {
        server.Bind(endPoint);
        server.Listen(1);

        var ex = Assert.Throws<ImapConnectionException>(() => {
          var serverEndPoint = server.LocalEndPoint as IPEndPoint;

          using (var session = new ImapSession(serverEndPoint.Address.ToString(), serverEndPoint.Port, 1000)) {
          }
        });

#if NETFRAMEWORK || NETSTANDARD2_0
        server.Close();
#else
        server.Dispose();
#endif

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      }
    }

    [Test]
    public void TestConnect_ConnectionTimeout_ServerNotAccepting()
    {
      ImapPseudoServer.StartNotAccepting((server) => {
        var ex = Assert.Throws<ImapConnectionException>(() => {
          using (var session = new ImapSession(server.Host, server.Port, 500)) {
          }
        });

        Assert.IsInstanceOf<TimeoutException>(ex.InnerException);
      });
    }

    [Test]
    public void TestConnect_SocketError()
    {
      var ex = Assert.Throws<ImapConnectionException>(() => {
        using (var session = new ImapSession("imap.invalid", 10143)) {
        }
      });

      Assert.IsInstanceOf<System.Net.Sockets.SocketException>(ex.InnerException);
    }

    [Test]
    public void TestConnect_Ssl()
    {
      ImapPseudoServer.Start((server) => {
        var streamUpgraded = false;

        using (var session = new ImapSession(server.Host,
                                             server.Port,
                                             createAuthenticatedStreamCallback: (connection, baseStream) => { streamUpgraded = true; return baseStream; /* TODO: return SslStream */})) {
          Assert.IsFalse(session.IsIdling);
          Assert.IsNull(session.SelectedMailbox);
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(string.Format("imaps://{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
          Assert.IsTrue(streamUpgraded, "stream upgraded");
          Assert.IsTrue(session.IsSecureConnection);
        }
      });
    }

    [Test]
    public void TestConnect_Ssl_ExceptionWhileUpgrading()
    {
      ImapPseudoServer.Start((server) => {
        try {
          using (var session = new ImapSession(server.Host,
                                               server.Port,
                                               createAuthenticatedStreamCallback: (connection, baseStream) => throw new System.Security.Authentication.AuthenticationException())) {
            Assert.Fail("connected");
          }
        }
        catch (ImapSecureConnectionException ex) {
          var upgradeException = ex.InnerException as ImapUpgradeConnectionException;

          Assert.IsNotNull(upgradeException);
          Assert.IsNotNull(upgradeException.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(upgradeException.InnerException);
        }
      });
    }

    [Test]
    public void TestDisconnect()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.Disconnect(false);

        Assert.IsTrue(session.IsDisposed);
        Assert.IsFalse(session.IsTransactionProceeding);

        Assert.Throws<ObjectDisposedException>(() => session.NoOp());

        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("State = {0}", session.State); });
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("IsIdling = {0}", session.IsIdling); });
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("SendTimeout = {0}", session.SendTimeout); });
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("ReceiveTimeout = {0}", session.ReceiveTimeout); } );
        Assert.Throws<ObjectDisposedException>(() => { Assert.Fail("TransactionTimeout = {0}", session.TransactionTimeout); } );
      });
    }

    [Test]
    public void TestImap4Rev1Incapable()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY NONE\r\n" +
                                     "$tag OK ImapCapability completed\r\n");

        ImapCapabilitySet caps;

        Assert.IsTrue((bool)session.Capability(out caps));
        Assert.IsNotNull(caps);
        Assert.IsFalse(caps.Contains(ImapCapability.Imap4Rev1));

        try {
          session.HandlesIncapableAsException = true;
          session.NoOp();
          Assert.Fail("exception not occured");
        }
        catch (ImapIncapableException) {
        }
      });
    }

    [Test]
    public void TestGreeting_WithCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 AUTH=PLAIN CHILDREN THREAD=REFERENCES X-EXTENSION1 X-EXTENSION2] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(7, session.ServerCapabilities.Count);

          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Children));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.CRAMMD5));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

          Assert.IsTrue(session.ServerCapabilities.Contains("THREAD=REFERENCES"));
          Assert.IsTrue(session.ServerCapabilities.Contains("X-EXTENSION1"));
          Assert.IsTrue(session.ServerCapabilities.Contains("X-EXTENSION2"));

          Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestGreeting_WithNoCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(0, session.ServerCapabilities.Count);

          Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

          Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestGreeting_PreAuth_WithCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* PREAUTH [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(2, session.ServerCapabilities.Count);
          Assert.AreEqual(string.Format("imap://{0}:{1}/", server.Host, server.Port), session.Authority.AbsoluteUri);

          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

          Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(ImapCapability.Imap4Rev1));
        }
      });
    }

    [Test]
    public void TestGreeting_WithNoImapCapability()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsTrue(!session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

          session.HandlesIncapableAsException = true;

          Assert.Throws<ImapIncapableException>(() => session.Login(new NetworkCredential("user", "pass")));
        }
      });
    }

    [Test]
    public void TestGreeting_HomeServerReferral_AsException()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n",
                                   null, null, true);

        try {
          using (var session = new ImapSession(server.Host, server.Port, handlesReferralAsException: true)) {
            Assert.Fail("connected without exception");
          }
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Server not accepting connections.  Try SERVER2"));
          StringAssert.AreEqualIgnoringCase("IMAP://user;AUTH=*@SERVER2/", ex.ReferToUri.AbsoluteUri);
        }
      });
    }

    [Test]
    public void TestGreeting_HomeServerReferral_AsError()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n",
                                   null, null, true);

        Assert.Throws<ImapConnectionException>(delegate {
          using (var session = new ImapSession(server.Host, server.Port)) {
          }
        });
      });
    }

    [Test]
    public void TestTransactionException_TransactionTimeout()
    {
      TestCases.WithConnectedSession((session, server) => {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK"); // not respond NOOP response termination(CRLF)

        Assert.Throws<TimeoutException>(() => session.NoOp());

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);

        Assert.Throws<ImapProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestTransactionException_SocketTimeout()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond NOOP
        //server.EnqueueTaggedResponse("$tag OK\r\n");

        Assert.Throws<TimeoutException>(() => session.NoOp());

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);

        Assert.Throws<ImapProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestTransactionException_InternalError()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "*\r\n" + // invalid response
                                     "$tag CAPABILITY Ok\r\n");

        var ex = Assert.Throws<ImapException>(() => session.Capability());

        Assert.IsNotNull(ex.InnerException);

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);

        Assert.Throws<ImapProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestDiconnectedFromServer_WithByeResponse()
    {
      TestCases.WithConnectedSession((session, server) => {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

        server.EnqueueUntaggedByeResponse("* BYE Disconnected for inactivity.\r\n");

        Assert.IsTrue(session.NoOp().Code == ImapCommandResultCode.Bye);

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);

        Assert.Throws<ImapProtocolViolationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestDiconnectedFromServer_WithoutByeResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [ALERT] will soon shutdown",
                                   null, null, true);

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.Throws<ImapConnectionException>(() => session.NoOp());

          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
          Assert.AreEqual(null, session.Authority);

          Assert.Throws<ImapProtocolViolationException>(() => session.NoOp());
        }
      });
    }

    [Test]
    public void TestReceivedCommandTagUnmatched()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("LOGIN",
                               "0001 OK authenticated\r\n");

        var ex = Assert.Throws<ImapException>(() => session.Login(new NetworkCredential("user", "pass")));

        Assert.IsNotNull(ex.InnerException);
        Assert.IsInstanceOf<ImapMalformedResponseException>(ex.InnerException);
      });
    }

    [Test]
    public void TestListStatusSelect()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK authenticated\r\n");

        Assert.IsTrue((bool)session.Login(new NetworkCredential("user", "pass")));

        var expectedAuthority = string.Format("imap://user@{0}/", server.HostPort);

        Assert.AreEqual(expectedAuthority, session.Authority.AbsoluteUri);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        session.Capability();

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                     "* LIST (\\HasChildren) \".\" \"INBOX\"\r\n" +
                                     "$tag OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.List("*", out mailboxes));

        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.HasChildren));
        Assert.AreEqual(expectedAuthority + "INBOX", mailboxes[0].Url.AbsoluteUri);

        // LSUB transaction
        server.EnqueueTaggedResponse("LSUB \"\" *\r\n",
                                     "* LSUB (\\HasChildren) \"/\" \"Trash\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \"/\" \"Trash/mail\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \".\" \"INBOX.&U,BTFw-\"\r\n" +
                                     "$tag OK LSUB completed\r\n");

        Assert.IsTrue((bool)session.Lsub("*", out mailboxes));

        Assert.AreEqual(4, mailboxes.Length);
        Assert.AreEqual("Trash", mailboxes[0].Name);
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.HasChildren));

        Assert.AreEqual("Trash/mail", mailboxes[1].Name);
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.HasNoChildren));
        Assert.AreEqual(expectedAuthority + "Trash/mail", mailboxes[1].Url.AbsoluteUri);

        Assert.AreEqual("INBOX.日本語", mailboxes[2].Name);
        Assert.AreEqual("INBOX.台北", mailboxes[3].Name);

        // STATUS transaction
        server.EnqueueTaggedResponse("STATUS Trash (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                     "* STATUS \"Trash\" (MESSAGES 3 RECENT 1 UIDVALIDITY 123456 UNSEEN 2 UIDNEXT 4)\r\n" +
                                     "$tag OK STATUS completed\r\n");

        Assert.IsTrue((bool)session.Status(mailboxes[0], ImapStatusDataItem.StandardAll));

        Assert.AreEqual(3, mailboxes[0].ExistsMessage);
        Assert.AreEqual(1, mailboxes[0].RecentMessage);
        Assert.AreEqual(123456, mailboxes[0].UidValidity);
        Assert.AreEqual(2, mailboxes[0].UnseenMessage);
        Assert.AreEqual(4, mailboxes[0].UidNext);

        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);

        server.EnqueueTaggedResponse("SELECT INBOX.&ZeVnLIqe-\r\n",
                                     "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                                     "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                                     "* 15 EXISTS\r\n" +
                                     "* 2 RECENT\r\n" +
                                     "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                                     "* OK [UIDNEXT 16]\r\n" +
                                     "* OK [UNSEEN 13]\r\n" +
                                     "$tag OK [READ-WRITE] SELECT completed\r\n");

        var mailbox = mailboxes[2];

        Assert.IsTrue((bool)session.Select(mailbox));

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.AreSame(mailbox, session.SelectedMailbox);

        Assert.AreEqual("INBOX.日本語", mailbox.Name);
        Assert.AreEqual(15, mailbox.ExistsMessage);
        Assert.AreEqual(2, mailbox.RecentMessage);
        Assert.AreEqual(0, mailbox.UnseenMessage);
        Assert.AreEqual(13, mailbox.FirstUnseen);
        Assert.AreEqual(1202674433, mailbox.UidValidity);
        Assert.AreEqual(16, mailbox.UidNext);
        Assert.IsFalse(mailbox.ReadOnly);
        Assert.IsTrue(mailbox.UidPersistent);
        Assert.AreEqual((new Uri(expectedAuthority + "INBOX.日本語;UIDVALIDITY=1202674433")).AbsoluteUri,
                        mailbox.Url.AbsoluteUri);

        /*
        Assert.AreEqual(selectedMailbox.Name, mailbox.Name);
        Assert.AreEqual(selectedMailbox.ExistsMessage, mailbox.ExistsMessage);
        Assert.AreEqual(selectedMailbox.RecentMessage, mailbox.RecentMessage);
        Assert.AreEqual(selectedMailbox.UnseenMessage, mailbox.UnseenMessage);
        Assert.AreEqual(selectedMailbox.FirstUnseen, mailbox.FirstUnseen);
        Assert.AreEqual(selectedMailbox.UidValidity, mailbox.UidValidity);
        Assert.AreEqual(selectedMailbox.UidNext, mailbox.UidNext);
        Assert.AreEqual(selectedMailbox.ReadOnly, mailbox.ReadOnly);
        Assert.AreEqual(selectedMailbox.ApplicableFlags.Count, mailbox.ApplicableFlags.Count);
        Assert.AreEqual(selectedMailbox.PermanentFlags.Count, mailbox.PermanentFlags.Count);
        */

        Assert.AreEqual(7, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("custom1"));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("custom2"));

        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.PermanentFlags.Contains("custom1"));

        // CLOSE transaction
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK CLOSE completed\r\n");

        Assert.IsTrue((bool)session.Close());

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      });
    }

    [Test]
    public void TestSerializeBinary_CommandResultCodeOk()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        var result = session.Capability();

        TestUtils.Assert.IsSerializableBinaryFormat(result, delegate(ImapCommandResult deserialized) {
          Assert.AreEqual(ImapCommandResultCode.Ok, deserialized.Code);
          Assert.IsNotNull(deserialized.TaggedStatusResponse);
          Assert.AreEqual("0002", deserialized.TaggedStatusResponse.Tag);
          Assert.AreEqual(ImapResponseCondition.Ok, deserialized.TaggedStatusResponse.Condition);
          Assert.IsNull(deserialized.TaggedStatusResponse.ResponseText.Code);
          Assert.AreEqual("CAPABILITY completed", deserialized.ResponseText);
          Assert.IsNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(2, responses.Length);
          Assert.IsInstanceOf<ImapDataResponse>(responses[0]);
          Assert.IsInstanceOf<ImapTaggedStatusResponse>(responses[1]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_CommandResultCodeRequestDone()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        var result = session.Login(new NetworkCredential("user", "pass")); // already authenticated

        TestUtils.Assert.IsSerializableBinaryFormat(result, delegate(ImapCommandResult deserialized) {
          Assert.AreEqual(ImapCommandResultCode.RequestDone, deserialized.Code);
          Assert.IsNull(deserialized.TaggedStatusResponse);
          Assert.IsNull(deserialized.ResponseText);
          Assert.IsNotNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(0, responses.Length);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_CommandResultCodeRequestError()
    {
      TestCases.WithConnectedSession(delegate(ImapSession session) {
        var result = session.Login(new NullCredential());

        TestUtils.Assert.IsSerializableBinaryFormat(result, delegate(ImapCommandResult deserialized) {
          Assert.AreEqual(ImapCommandResultCode.RequestError, deserialized.Code);
          Assert.IsNull(deserialized.TaggedStatusResponse);
          Assert.IsNull(deserialized.ResponseText);
          Assert.IsNotNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(0, responses.Length);
        });
      });
    }

    [Test]
    public void TestGenericCommand_WithoutArguments()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        // generic command transaction
        server.EnqueueTaggedResponse("X-EXT-COMMAND\r\n",
                                     "* X-EXT-COMMAND data\r\n" +
                                     "$tag OK X-EXT-COMMAND completed\r\n");

        ImapDataResponse[] dataResps;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps));

        Assert.AreEqual(1, dataResps.Length);
        Assert.AreEqual(2, dataResps[0].Data.Length);
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[0].Format);
        Assert.AreEqual("X-EXT-COMMAND", dataResps[0].Data[0].GetTextAsString());
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[1].Format);
        Assert.AreEqual("data", dataResps[0].Data[1].GetTextAsString());
      });
    }

    [Test]
    public void TestGenericCommand_WithArguments()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        // generic command transaction
        server.EnqueueTaggedResponse("X-EXT-COMMAND ",
                                     "$tag OK X-EXT-COMMAND completed\r\n");

        ImapDataResponse[] dataResps;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps,
                                           new ImapParenthesizedString("arg1",
                                                                       "arg2",
                                                                       new ImapParenthesizedString("arg3-1", "arg3-2"),
                                                                       new ImapLiteralString("ascii text", OctetEncoding.EightBits)),
                                           new ImapParenthesizedString(new ImapQuotedString("quoted"),
                                                                       "nonquoted")
                                           ));

        Assert.AreEqual("0002 X-EXT-COMMAND (arg1 arg2 (arg3-1 arg3-2) {10}\r\n" +
                        "ascii text) (\"quoted\" nonquoted)\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(0, dataResps.Length);
      });
    }

    [Test]
    public void TestCommand_Tag()
    {
      TestCases.WithConnectedSession((session, server) => {
        server.EnqueueResponse("0000 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0000 NOOP\r\n", server.DequeueRequest());

        server.EnqueueResponse("0001 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0001 NOOP\r\n", server.DequeueRequest());

        server.EnqueueResponse("0002 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0002 NOOP\r\n", server.DequeueRequest());
      });
    }

#if !DEBUG
    [Test]
    public void TestCommandTagNotOverflow()
    {
      const int div = 1 << 8;
      const int commandTagMax = 0x10000;

      TestCases.WithConnectedSession((session, server) => {
        for (var i = 0; i < (commandTagMax / div); i++) {
          for (var j = 0; j < div; j++) {
            server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
          }

          for (var j = 0; j < div; j++) {
            Assert.IsTrue((bool)session.NoOp());
          }

          for (var j = 0; j < div; j++) {
            server.DequeueRequest();
          }
        }

        server.EnqueueResponse("NOOP\r\n",
                               "0000 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0000 NOOP\r\n", server.DequeueRequest());
      });
    }
#endif
  }
}
