using System;
using System.Collections.Generic;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionCommandsAuthenticatedOrSelectedStateTests {
    [Test]
    public void TestNamespace()
    {
      TestCases.WithAuthenticatedSession(new[] {"NAMESPACE"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Namespace));

        // NAMESPACE transaction
        server.EnqueueTaggedResponse("NAMESPACE\r\n",
                                     "* NAMESPACE ((\"\" \"/\")) ((\"~\" \"/\")) ((\"#shared/\" \"/\")(\"#public/\" \"/\")(\"#ftp/\" \"/\")(\"#news.\" \".\"))\r\n" +
                                     "$tag OK NAMESPACE command completed\r\n");

        ImapNamespace namespaces = null;

        Assert.IsTrue((bool)session.Namespace(out namespaces));

        Assert.IsNotNull(namespaces);
        Assert.AreEqual(1, namespaces.PersonalNamespaces.Count);
        Assert.AreEqual(string.Empty, namespaces.PersonalNamespaces[0].Prefix);
        Assert.AreEqual("/", namespaces.PersonalNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, namespaces.PersonalNamespaces[0].Extensions.Count);

        Assert.AreEqual(1, namespaces.OtherUsersNamespaces.Count);
        Assert.AreEqual("~", namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.AreEqual("/", namespaces.OtherUsersNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, namespaces.OtherUsersNamespaces[0].Extensions.Count);

        Assert.AreEqual(4, namespaces.SharedNamespaces.Count);
        Assert.AreEqual("#shared/", namespaces.SharedNamespaces[0].Prefix);
        Assert.AreEqual("#public/", namespaces.SharedNamespaces[1].Prefix);
        Assert.AreEqual("#ftp/", namespaces.SharedNamespaces[2].Prefix);
        Assert.AreEqual("#news.", namespaces.SharedNamespaces[3].Prefix);

        Assert.IsTrue(namespaces.PersonalNamespaces.Count == session.Namespaces.PersonalNamespaces.Count);
        Assert.IsTrue(namespaces.OtherUsersNamespaces.Count == session.Namespaces.OtherUsersNamespaces.Count);
        Assert.IsTrue(namespaces.SharedNamespaces.Count == session.Namespaces.SharedNamespaces.Count);
        Assert.IsTrue(namespaces.PersonalNamespaces[0].Prefix == session.Namespaces.PersonalNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.OtherUsersNamespaces[0].Prefix == session.Namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[0].Prefix == session.Namespaces.SharedNamespaces[0].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[1].Prefix == session.Namespaces.SharedNamespaces[1].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[2].Prefix == session.Namespaces.SharedNamespaces[2].Prefix);
        Assert.IsTrue(namespaces.SharedNamespaces[3].Prefix == session.Namespaces.SharedNamespaces[3].Prefix);
      });
    }

    [Test]
    public void TestNamespace_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Namespace));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.Namespace());
      });
    }

    [Test]
    public void TestComparator_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.I18NLevel2));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.Comparator());
      });
    }

    [Test]
    public void TestComparator()
    {
      TestCases.WithAuthenticatedSession(new[] {"I18NLEVEL=2"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.I18NLevel2));
        Assert.AreEqual(ImapCollationAlgorithm.Default, session.ActiveComparator);

        // COMPARATOR transaction
        server.EnqueueTaggedResponse("COMPARATOR\r\n",
                                     "* COMPARATOR i;unicode-casemap\r\n" +
                                     "$tag OK Will use i;unicode-casemap for collation\r\n");

        ImapCollationAlgorithm activeComparator;

        Assert.IsTrue((bool)session.Comparator(out activeComparator));

        Assert.AreEqual(ImapCollationAlgorithm.UnicodeCasemap, session.ActiveComparator);
      });
    }

    [Test]
    public void TestComparator_ChangeComparator()
    {
      TestCases.WithAuthenticatedSession(new[] {"I18NLEVEL=2"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.I18NLevel2));
        Assert.AreEqual(ImapCollationAlgorithm.Default, session.ActiveComparator);

        // COMPARATOR transaction
        server.EnqueueTaggedResponse("COMPARATOR \"cz;*\" \"i;basic\"\r\n",
                                     "* COMPARATOR i;basic\r\n" +
                                     "$tag OK Will use i;basic for collation\r\n");

        ImapCollationAlgorithm activeComparator;
        ImapCollationAlgorithm[] matchingComparators;

        Assert.IsTrue((bool)session.Comparator(out activeComparator,
                                               out matchingComparators,
                                               new ImapCollationAlgorithm("cz;*"),
                                               new ImapCollationAlgorithm("i;basic")));

        Assert.AreEqual(new ImapCollationAlgorithm("i;basic"), activeComparator);
        Assert.AreEqual(0, matchingComparators.Length);

        Assert.AreEqual(new ImapCollationAlgorithm("i;basic"), session.ActiveComparator);
      });
    }

    [Test]
    public void TestSetQuota()
    {
      TestCases.WithAuthenticatedSession(new[] {"QUOTA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Quota));

        // SETQUOTA transaction
        server.EnqueueTaggedResponse("SETQUOTA \"\" (STORAGE 512)\r\n",
                                     "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                     "$tag OK Setquota completed\r\n");

        ImapQuota changedQuota;

        Assert.IsTrue((bool)session.SetQuota(string.Empty, "STORAGE", 512L, out changedQuota));

        Assert.IsNotNull(changedQuota);
        Assert.AreEqual(string.Empty, changedQuota.Root);
        Assert.AreEqual(1, changedQuota.Resources.Count);
        Assert.AreEqual("STORAGE", changedQuota.Resources[0].Name);
        Assert.AreEqual(10L, changedQuota.Resources[0].Usage);
        Assert.AreEqual(512L, changedQuota.Resources[0].Limit);
      });
    }

    [Test]
    public void TestSetQuota_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Quota));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.SetQuota(string.Empty, "STORAGE", 512L));
      });
    }

    [Test]
    public void TestGetQuota()
    {
      TestCases.WithAuthenticatedSession(new[] {"QUOTA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Quota));

        // SETQUOTA transaction
        server.EnqueueTaggedResponse("GETQUOTA \"\"\r\n",
                                     "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                     "$tag OK Getquota completed\r\n");

        ImapQuota quota;

        Assert.IsTrue((bool)session.GetQuota(string.Empty, out quota));

        Assert.IsNotNull(quota);
        Assert.AreEqual(string.Empty, quota.Root);
        Assert.AreEqual(1, quota.Resources.Count);
        Assert.AreEqual("STORAGE", quota.Resources[0].Name);
        Assert.AreEqual(10L, quota.Resources[0].Usage);
        Assert.AreEqual(512L, quota.Resources[0].Limit);
      });
    }

    [Test]
    public void TestGetQuota_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Quota));

        session.HandlesIncapableAsException = true;

        ImapQuota quota;

        Assert.Throws<ImapIncapableException>(() => session.GetQuota(string.Empty, out quota));
      });
    }

    [Test]
    public void TestGetQuotaRoot()
    {
      TestCases.WithAuthenticatedSession(new[] {"QUOTA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Quota));

        // GETQUOTAROOT transaction
        server.EnqueueTaggedResponse("GETQUOTAROOT INBOX\r\n",
                                     "* QUOTAROOT INBOX \"\"\r\n" +
                                     "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                     "$tag OK Getquota completed\r\n");

        IReadOnlyDictionary<string, ImapQuota[]> quotaRoots;

        Assert.IsTrue((bool)session.GetQuotaRoot("INBOX", out quotaRoots));

        Assert.IsNotNull(quotaRoots);
        Assert.AreEqual(1, quotaRoots.Count);
        Assert.IsTrue(quotaRoots.ContainsKey("INBOX"));

        var inboxQuotaRoot = quotaRoots["INBOX"];

        Assert.IsNotNull(inboxQuotaRoot);
        Assert.AreEqual(1, inboxQuotaRoot.Length);
        Assert.AreEqual(string.Empty, inboxQuotaRoot[0].Root);
        Assert.AreEqual(1, inboxQuotaRoot[0].Resources.Count);
        Assert.AreEqual("STORAGE", inboxQuotaRoot[0].Resources[0].Name);
        Assert.AreEqual(10L, inboxQuotaRoot[0].Resources[0].Usage);
        Assert.AreEqual(512L, inboxQuotaRoot[0].Resources[0].Limit);
      });
    }

    [Test]
    public void TestGetQuotaRoot_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Quota));

        session.HandlesIncapableAsException = true;

        IReadOnlyDictionary<string, ImapQuota[]> quotaRoots;

        Assert.Throws<ImapIncapableException>(() => session.GetQuotaRoot("INBOX", out quotaRoots));
      });
    }

    [Test]
    public void TestGetMetadata_Server()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA-SERVER"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("GETMETADATA \"\" /shared/comment\r\n",
                                     "* METADATA \"\" (/shared/comment \"Shared comment\")\r\n" +
                                     "$tag OK GETMETADATA complete\r\n");

        ImapMetadata[] metadata;

        Assert.IsTrue((bool)session.GetMetadata("/shared/comment", out metadata));

        Assert.IsNotNull(metadata);
        Assert.AreEqual(1, metadata.Length);
        Assert.AreEqual("/shared/comment", metadata[0].EntryName);
        //Assert.AreEqual("Shared comment", (string)metadata[0].Value);
        Assert.IsTrue(metadata[0].Value.Equals("Shared comment"));
      });
    }

    [Test]
    public void TestGetMetadata_Server_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        ImapMetadata[] metadata;

        Assert.Throws<ImapIncapableException>(() => session.GetMetadata("/shared/comment", out metadata));
      });
    }

    [Test]
    public void TestGetMetadata_Server_Capable_1()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        ImapMetadata[] metadata;

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("* METADATA \"\" (/shared/comment \"Shared comment\")\r\n" +
                                     "$tag OK GETMETADATA complete\r\n");

        Assert.IsTrue((bool)session.GetMetadata("/shared/comment", out metadata));
      });
    }

    [Test]
    public void TestGetMetadata_Server_Capable_2()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA-SERVER"}, (session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        ImapMetadata[] metadata;

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("* METADATA \"\" (/shared/comment \"Shared comment\")\r\n" +
                                     "$tag OK GETMETADATA complete\r\n");

        Assert.IsTrue((bool)session.GetMetadata("/shared/comment", out metadata));
      });
    }

    [Test]
    public void TestGetMetadata_1()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("GETMETADATA INBOX /private/comment\r\n",
                                     "* METADATA INBOX (/private/comment \"My own comment\")\r\n" +
                                     "$tag OK GETMETADATA complete\r\n");

        ImapMetadata[] metadata;

        Assert.IsTrue((bool)session.GetMetadata("INBOX", "/private/comment", out metadata));

        Assert.IsNotNull(metadata);
        Assert.AreEqual(1, metadata.Length);
        Assert.AreEqual("/private/comment", metadata[0].EntryName);
        //Assert.AreEqual("My own comment", (string)metadata[0].Value);
        Assert.IsTrue(metadata[0].Value.Equals("My own comment"));
      });
    }

    [Test]
    public void TestGetMetadata_2()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("GETMETADATA INBOX (/shared/comment /private/comment)\r\n",
                                     "* METADATA INBOX (/shared/comment \"Shared comment\" " +
                                     "/private/comment \"My own comment\")\r\n"+
                                     "$tag OK GETMETADATA complete\r\n");

        ImapMetadata[] metadata;

        Assert.IsTrue((bool)session.GetMetadata("INBOX", new[] {"/shared/comment", "/private/comment"}, out metadata));

        Assert.IsNotNull(metadata);
        Assert.AreEqual(2, metadata.Length);
        Assert.AreEqual("/shared/comment", metadata[0].EntryName);
        //Assert.AreEqual("Shared comment", (string)metadata[0].Value);
        Assert.IsTrue(metadata[0].Value.Equals("Shared comment"));
        Assert.AreEqual("/private/comment", metadata[1].EntryName);
        //Assert.AreEqual("My own comment", (string)metadata[1].Value);
        Assert.IsTrue(metadata[1].Value.Equals("My own comment"));
      });
    }

    [Test]
    public void TestGetMetadata_MaxSize()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("GETMETADATA INBOX (MAXSIZE 1024) (/shared/comment /private/comment)\r\n",
                                     "* METADATA INBOX (/private/comment \"My own comment\")\r\n"+
                                     "$tag OK [METADATA LONGENTRIES 2199] GETMETADATA complete\r\n");

        ImapMetadata[] metadata;

        var result = session.GetMetadata("INBOX",
                                         new[] {"/shared/comment", "/private/comment"},
                                         ImapGetMetadataOptions.MaxSize(1024L),
                                         out metadata);

        Assert.IsTrue((bool)result);

        Assert.IsNotNull(metadata);
        Assert.AreEqual(1, metadata.Length);
        Assert.AreEqual("/private/comment", metadata[0].EntryName);
        //Assert.AreEqual("My own comment", (string)metadata[0].Value);
        Assert.IsTrue(metadata[0].Value.Equals("My own comment"));

        var resp = result.GetResponseCode(ImapResponseCode.MetadataLongEntries);

        Assert.IsNotNull(resp);
        Assert.AreEqual(2199, ImapResponseTextConverter.FromMetadataLongEntries(resp.ResponseText));
      });
    }

    [Test]
    public void TestGetMetadata_Depth()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        // GETMETADATA transaction
        server.EnqueueTaggedResponse("GETMETADATA INBOX (DEPTH 1) /private/filters/values\r\n",
                                     "* METADATA INBOX (/private/filters/values/small " +
                                     "\"SMALLER 5000\" /private/filters/values/boss " +
                                     "\"FROM \\\"boss@example.com\\\"\")\r\n" +
                                     "$tag OK GETMETADATA complete\r\n");

        ImapMetadata[] metadata;

        Assert.IsTrue((bool)session.GetMetadata("INBOX",
                                                "/private/filters/values",
                                                ImapGetMetadataOptions.Depth1,
                                                out metadata));

        Assert.IsNotNull(metadata);
        Assert.AreEqual(2, metadata.Length);
        Assert.AreEqual("/private/filters/values/small", metadata[0].EntryName);
        //Assert.AreEqual("SMALLER 5000", (string)metadata[0].Value);
        Assert.IsTrue(metadata[0].Value.Equals("SMALLER 5000"));
        Assert.AreEqual("/private/filters/values/boss", metadata[1].EntryName);
        //Assert.AreEqual("FROM \"boss@example.com\"", (string)metadata[1].Value);
        Assert.IsTrue(metadata[1].Value.Equals("FROM \"boss@example.com\""));
      });
    }

    [Test]
    public void TestGetMetadata_Incapable_1()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        session.HandlesIncapableAsException = true;

        ImapMetadata[] metadata;

        Assert.Throws<ImapIncapableException>(() => session.GetMetadata("/shared/comment", out metadata));
      });
    }

    [Test]
    public void TestGetMetadata_Incapable_2()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA-SERVER"}, (session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        ImapMetadata[] metadata;

        Assert.Throws<ImapIncapableException>(() => session.GetMetadata("INBOX", "/shared/comment", out metadata));
      });
    }

    [Test]
    public void TestSetMetadata_Server()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA-SERVER"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        // SETMETADATA transaction
        server.EnqueueTaggedResponse("SETMETADATA \"\" (/shared/vendor/foo/bar NIL)\r\n",
                                     "$tag OK SETMETADATA complete\r\n");

        Assert.IsTrue((bool)session.SetMetadata(new[] {"/shared/vendor/foo/bar"}));
      });
    }

    [Test]
    public void TestSetMetadata_Server_Incapable()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.SetMetadata(new[] {"/shared/vendor/foo/bar"}));
      });
    }

    [Test]
    public void TestSetMetadata_Server_Capable_1()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        // SETMETADATA transaction
        server.EnqueueTaggedResponse("$tag OK SETMETADATA complete\r\n");

        Assert.IsTrue((bool)session.SetMetadata(new[] {"/shared/vendor/foo/bar"}));
      });
    }

    [Test]
    public void TestSetMetadata_Server_Capable_2()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA-SERVER"}, (session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        // SETMETADATA transaction
        server.EnqueueTaggedResponse("$tag OK SETMETADATA complete\r\n");

        Assert.IsTrue((bool)session.SetMetadata(new[] {"/shared/vendor/foo/bar"}));
      });
    }

    [Test]
    public void TestSetMetadata_1()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        server.EnqueueTaggedResponse("SETMETADATA INBOX (/private/comment ",
                                     "$tag OK SETMETADATA complete\r\n");

        var metadata = new ImapMetadata("/private/comment",
                                        new ImapLiteralString("My new comment across\n" +
                                                              "two lines.\n"));

        Assert.IsTrue((bool)session.SetMetadata("INBOX", metadata));

        Assert.AreEqual("0002 SETMETADATA INBOX (/private/comment {33}\r\n" +
                        "My new comment across\n" +
                        "two lines.\n" +
                        ")\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestSetMetadata_2()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        // SETMETADATA transaction
        server.EnqueueTaggedResponse("SETMETADATA INBOX (/private/comment NIL)\r\n",
                                     "$tag OK SETMETADATA complete\r\n");

        Assert.IsTrue((bool)session.SetMetadata("INBOX", "/private/comment"));
      });
    }

    [Test]
    public void TestSetMetadata_3()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA"}, (session, server) => {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        // SETMETADATA transaction
        server.EnqueueTaggedResponse("SETMETADATA INBOX (/private/comment ",
                                     "$tag OK SETMETADATA complete\r\n");

        var metadata = new[] {
          new ImapMetadata("/private/comment", new ImapQuotedString("My new comment")),
          new ImapMetadata("/shared/comment", new ImapQuotedString("This one is for you!")),
        };

        Assert.IsTrue((bool)session.SetMetadata("INBOX", metadata));

        Assert.AreEqual("0002 SETMETADATA INBOX (/private/comment \"My new comment\" " +
                        "/shared/comment \"This one is for you!\")\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestSetMetadata_Incapable_1()
    {
      TestCases.WithAuthenticatedSession((session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.SetMetadata("INBOX", "/private/comment"));
      });
    }

    [Test]
    public void TestSetMetadata_Incapable_2()
    {
      TestCases.WithAuthenticatedSession(new[] {"METADATA-SERVER"}, (session, server) => {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Metadata));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.MetadataServer));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.SetMetadata("INBOX", "/private/comment"));
      });
    }
  }
}
