using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoOperationsFetchTests {
    [SetUp]
    public void SetUp()
    {
      TestUtils.Encodings.RegisterCodePagesEncodingProvider();
    }

    [Test]
    public void TestMessagePart_StaticAttributeFetched()
    {
      var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
      var envelope =
        "(" +
        "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
        "\"test message\" " + 
        "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
        "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
        "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
        "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
        "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
        "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
        "\"<in.reply.to@mail.example.com>\" " + 
        "\"<message.id@mail.example.com>\" " + 
        ")";
      var fetchResp =
        "* FETCH 1 (UID 1 " + 
        "BODYSTRUCTURE " + bodystructure +
        "ENVELOPE " + envelope +
        "INTERNALDATE \"23-Mar-2010 11:22:33 +0900\" RFC822.SIZE 1024)\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp, ImapMessageFetchAttributeOptions.StaticAttributes,
                             (server, message) => {
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.MessagePart);
        Assert.AreSame(message.MessagePart, message.GetMessagePart(string.Empty));
        Assert.IsTrue(message.MessagePart.ContentType.EqualsIgnoreCase("text/plain"));
        Assert.AreEqual("7bit", message.MessagePart.ContentTransferEncoding);
      });
    }

    [Test]
    public void TestMessagePart_StaticAttributeNotFetched()
    {
      var fetchResp =
        "* FETCH 1 (UID 1)\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp,
                             ImapMessageFetchAttributeOptions.None,
                             (server, message) => {
        Assert.IsFalse(message.Client.IsBusy);

        var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
        var envelope =
          "(" +
          "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
          "\"test message\" " + 
          "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
          "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
          "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
          "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
          "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
          "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
          "\"<in.reply.to@mail.example.com>\" " + 
          "\"<message.id@mail.example.com>\" " + 
          ")";

        server.EnqueueTaggedResponse("UID FETCH 1",
                                     "* FETCH 1 (UID 1 " + 
                                     "BODYSTRUCTURE " + bodystructure +
                                     "ENVELOPE " + envelope +
                                     "INTERNALDATE \"23-Mar-2010 11:22:33 +0900\" RFC822.SIZE 1024)\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsNotNull(message.MessagePart);
        Assert.AreSame(message.MessagePart, message.GetMessagePart(string.Empty));
        Assert.IsTrue(message.MessagePart.ContentType.EqualsIgnoreCase("text/plain"));
        Assert.AreEqual("7bit", message.MessagePart.ContentTransferEncoding);
      });
    }

    private void TestFetchMultipartBodystructure(Action<ImapMessageInfo> action)
    {
      var envelope = "(" +
                     "\"Thu, 25 Mar 2010 13:43:30 +0900\" " +
                     "\"subject\" " + 
                     "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                     "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                     "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                     "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                     "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                     "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                     "\"<in.reply.to@mail.example.com>\" " + 
                     "\"<message.id@mail.example.com>\" " + 
                     ") ";
      var bodystructure = "(" +
                          "(\"text\" \"plain\" (\"charset\" \"ISO-2022-JP\") NIL NIL \"7bit\" 6 1 NIL NIL NIL NIL)" +
                          "(\"message\" \"rfc822\" (\"name\" \"=?ISO-2022-JP?B?GyRCRTpJVSVhJUMlOyE8JTgbKEI=?=\") " +
                          "NIL NIL \"7bit\" 188 (NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL) " +
                          "(\"text\" \"html\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" 121 6 NIL NIL NIL NIL) 10 NIL " +
                          "(\"inline\" (\"filename\" \"ISO-2022-JP''%1B%24%42%45%3A%49%55%25%61%25%43%25%3B%21%3C%25%38%1B%28%42\")) NIL NIL) " +
                          "\"mixed\" (\"boundary\" \"------------040401080108050302040809\") NIL NIL NIL" +
                          ") ";

      var fetchResp = "* FETCH 1 (UID 1 " + 
                      "BODYSTRUCTURE " + bodystructure +
                      "ENVELOPE " + envelope +
                      "INTERNALDATE \"23-Mar-2010 11:22:33 +0900\" RFC822.SIZE 1024)\r\n" +
                      "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp, ImapMessageFetchAttributeOptions.StaticAttributes,
                             (server, message) => {
        Assert.IsFalse(message.Client.IsBusy);

        action(message);
      });
    }

    [Test]
    public void TestGetMessagePart()
    {
      TestFetchMultipartBodystructure(delegate(ImapMessageInfo message) {
        Assert.AreSame(message.MessagePart, message.GetMessagePart((int[])null));
        Assert.AreSame(message.MessagePart, message.GetMessagePart(new int[0]));
        Assert.AreSame(message.MessagePart, message.GetMessagePart(Enumerable.Empty<int>()));

        Assert.AreEqual("text/plain", (string)message.GetMessagePart(1).ContentType);
        Assert.AreEqual("text/plain", (string)message.GetMessagePart("1").ContentType);
        Assert.AreEqual("message/rfc822", (string)message.GetMessagePart(2).ContentType);
        Assert.AreEqual("message/rfc822", (string)message.GetMessagePart("2").ContentType);
        Assert.AreEqual("text/html", (string)message.GetMessagePart(2, 1).ContentType);
        Assert.AreEqual("text/html", (string)message.GetMessagePart("2.1").ContentType);

        Assert.IsNull(message.GetMessagePart(2, 2));
        Assert.IsNull(message.GetMessagePart(3));
      });
    }

    [Test]
    public void TestGetMessagePart_SpecifierInString_Valid()
    {
      TestFetchMultipartBodystructure(delegate(ImapMessageInfo message) {
        Assert.DoesNotThrow(() => message.GetMessagePart("1"));
        Assert.DoesNotThrow(() => message.GetMessagePart("10"));
        Assert.DoesNotThrow(() => message.GetMessagePart("100"));
        Assert.DoesNotThrow(() => message.GetMessagePart("1.1"));
        Assert.DoesNotThrow(() => message.GetMessagePart("1.10"));
        Assert.DoesNotThrow(() => message.GetMessagePart("1.100"));
        Assert.DoesNotThrow(() => message.GetMessagePart("1.1.1"));
        Assert.DoesNotThrow(() => message.GetMessagePart("10.1.1"));
        Assert.DoesNotThrow(() => message.GetMessagePart("100.1.1"));

        Assert.DoesNotThrow(() => message.GetMessagePart("3.1"));
        Assert.DoesNotThrow(() => message.GetMessagePart("4"));
        Assert.DoesNotThrow(() => message.GetMessagePart("4.2.2.2"));
        Assert.DoesNotThrow(() => message.GetMessagePart("9"));
        Assert.DoesNotThrow(() => message.GetMessagePart("9.9"));
      });
    }

    [Test]
    public void TestGetStructureOf_SpecifierInString_Invalid()
    {
      TestFetchMultipartBodystructure(delegate(ImapMessageInfo message) {
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("0"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("01"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("001"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("."));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart(".1"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("1.01"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("1.001"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("01.1"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("001.1"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("1.1."));

        Assert.Throws<ArgumentException>(() => message.GetMessagePart("１"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("1１"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("1.１"));
        Assert.Throws<ArgumentException>(() => message.GetMessagePart("1.1１"));

        Assert.Throws<ArgumentException>(() => message.GetMessagePart("A"));
      });
    }

    [Test]
    public void TestFindMessagePart()
    {
      TestFetchMultipartBodystructure(delegate(ImapMessageInfo message) {
        Assert.AreEqual("1", (string)message.FindMessagePart(new MimeType("text/plain")).Specifier);
        Assert.AreEqual("2", (string)message.FindMessagePart(new MimeType("message/rfc822")).Specifier);
        Assert.AreEqual("2.1", (string)message.FindMessagePart(new MimeType("text/html")).Specifier);

        Assert.IsNull(message.FindMessagePart(new MimeType("image/jpeg")));
      });
    }

    private void TestFetchMessage(Action<ImapPseudoServer, ImapMessageInfo, string, int> action)
    {
      TestFetchMessage(false, string.Empty, action);
    }

    private void TestFetchMessage(bool setSeen,
                                  string initialFlags,
                                  Action<ImapPseudoServer, ImapMessageInfo, string, int> action)
    {
      const string messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
";

      TestCases.WithOpenedMailbox("INBOX",
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                     string.Format("* FETCH 1 (UID 1 FLAGS ({0}))\r\n", initialFlags ?? string.Empty) +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        server.DequeueRequest();

        var octets = OctetEncoding.EightBits.GetByteCount(messageBody);

        server.EnqueueTaggedResponse(string.Format("* {0} FETCH (RFC822.SIZE {1} {3}BODY[] {{{1}}}\r\n{2})\r\n",
                                                   message.Uid,
                                                   octets,
                                                   messageBody,
                                                   setSeen ? "FLAGS (\\Seen) " : string.Empty) +
                                     "$tag OK done\r\n");

        Assert.IsFalse(message.Client.IsBusy);

        action(server, message, messageBody, octets);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead()
    {
      TestFetchMessage(false,
                       null,
                       (server, message, expectedMessageBody, expectedOctets) => {
        Assert.IsFalse(message.IsSeen);

        using (var stream = message.OpenRead()) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);
          Assert.AreEqual(expectedMessageBody,
                          (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());

          Assert.IsFalse(message.Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsFalse(message.IsSeen);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");

          Assert.Throws<InvalidOperationException>(() => message.Client.Refresh());
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenRead_CloseStream()
    {
      TestFetchMessage(false,
                       null,
                       (server, message, expectedMessageBody, expectedOctets) => {
        Assert.IsFalse(message.IsSeen);

        using (var stream = message.OpenRead()) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");

          Assert.Throws<InvalidOperationException>(() => message.Client.Refresh());

          stream.Dispose();

          Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");

          Assert.IsFalse(stream.CanRead, "CanRead");
          Assert.IsFalse(stream.CanWrite, "CanWrite");
          Assert.IsFalse(stream.CanSeek, "CanSeek");
          Assert.IsTrue(stream.CanTimeout, "CanTimeout");

          Assert.Throws<ObjectDisposedException>(() => stream.ReadByte());

          stream.Dispose();

          Assert.IsFalse(message.Client.IsBusy, "IsBusy; after re-close");
        }
      });
    }

    [Test]
    public void TestOpenRead_SetSeen()
    {
      TestFetchMessage(true,
                       null,
                       (server, message, expectedMessageBody, expectedOctets) => {
        Assert.IsFalse(message.IsSeen);

        using (var stream = message.OpenRead(ImapMessageFetchBodyOptions.SetSeen)) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);
          Assert.AreEqual(expectedMessageBody,
                          (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());

          Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(message.IsSeen);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenRead_FlagsNotChange()
    {
      TestFetchMessage(false,
                       "\\Seen $label1",
                       (server, message, expectedMessageBody, expectedOctets) => {
        Assert.AreEqual(2, message.Flags.Count);
        Assert.IsTrue(message.IsSeen);
        Assert.IsTrue(message.Flags.Contains("$label1"));

        using (var stream = message.OpenRead(ImapMessageFetchBodyOptions.SetSeen)) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);
          Assert.AreEqual(expectedMessageBody,
                          (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());

          Assert.AreEqual(2, message.Flags.Count);
          Assert.IsTrue(message.IsSeen);
          Assert.IsTrue(message.Flags.Contains("$label1"));

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenRead_DecodeContent()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<NotSupportedException>(() => message.OpenRead(ImapMessageFetchBodyOptions.DecodeContent));
        Assert.IsFalse(message.Client.IsBusy);
      });

    }

    [Test]
    public void TestOpenRead_OmitHeader()
    {
      TestFetchMessage(false,
                       null,
                       (server, message, expectedMessageBody, expectedOctets) => {
        Assert.IsFalse(message.IsSeen);

        using (var stream = message.OpenRead(ImapMessageFetchBodyOptions.OmitHeader)) {
          StringAssert.EndsWith("UID FETCH 1 (BODY.PEEK[TEXT]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(0L, stream.Length);

          Assert.IsFalse(message.Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsFalse(message.IsSeen);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenRead_RangedFetch()
    {
      TestFetchMessage(false,
                       null,
                       (server, message, expectedMessageBody, expectedOctets) => {
        Assert.IsFalse(message.IsSeen);

        var fetchRange = new ImapPartialRange(128, 64);

        using (var stream = message.OpenRead(fetchRange)) {
          StringAssert.EndsWith("UID FETCH 1 (BODY.PEEK[]<128.64>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(fetchRange.Length, stream.Length);

          Assert.IsFalse(message.Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsFalse(message.IsSeen);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenRead_MailboxClosed()
    {
      TestCases.WithMessageMailboxClosed((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => message.OpenRead());

        TestUtils.Assert.IsSerializableBinaryFormat(ex);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_VanishedMessage()
    {
      TestCases.WithMessageVanished((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => message.OpenRead());

        TestUtils.Assert.IsSerializableBinaryFormat(ex);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_MessageVanishingByFetch()
    {
      TestCases.WithMessage((server, message) => {
        Assert.AreNotEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsFalse(message.IsDeleted);

        server.EnqueueTaggedResponse("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n",
                                     "$tag OK done\r\n");

        var ex = Assert.Throws<ImapMessageVanishedException>(() => message.OpenRead());

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(ex.VanishedMessage);
        Assert.AreSame(message, ex.VanishedMessage);

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapMessageVanishedException deserialized) {
          Assert.IsNull(deserialized.VanishedMessage);
        });

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsTrue(message.IsDeleted);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));
        Assert.IsFalse(message.IsAvailable);
      });
    }

#if false
    [Test, Ignore("this will never happen")]
    public void TestOpenRead_SequenceChanged()
    {
      const string messageBody = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

test message

3 uid fetch 6 (body[])
* 6 FETCH (UID 6 BODY[] {1476}
hogehogehoge
)
3 OK Fetch completed.
";

      var fetchResp =
        "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
        "* FETCH 2 (UID 2 FLAGS (\\Seen))\r\n" +
        "* FETCH 3 (UID 3 FLAGS (\\Recent))\r\n" +
        "$tag OK done\r\n";

      TestMessages(fetchResp, ImapMessageFetchAttributeOptions.DynamicAttributes,
                   delegate(ImapPseudoServer server, ImapMessageInfo[] messages) {
        Assert.IsFalse(messages[0].IsDeleted);
        Assert.IsFalse(messages[1].IsDeleted);
        Assert.IsFalse(messages[2].IsDeleted);

        // FETCH
        server.EnqueueTaggedResponse("* EXPUNGE 1\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     string.Format("* 1 FETCH (RFC822.SIZE {0} BODY[] {{{0}}}\r\n{1})\r\n",
                                                   OctetEncoding.EightBits.GetByteCount(messageBody),
                                                   messageBody) +
                                     "* 1 FETCH (FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var message = messages[2];

        Assert.IsFalse(message.IsSeen);
        Assert.IsTrue(message.IsRecent);

        using (var stream = message.OpenRead(ImapMessageFetchBodyOptions.SetSeen)) {
          Assert.IsNotNull(stream);

          StringAssert.EndsWith("UID FETCH 3 (RFC822.SIZE BODY[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsTrue(messages[0].IsDeleted);
          Assert.IsTrue(messages[1].IsDeleted);

          Assert.AreEqual(1L, message.Sequence);
          Assert.AreEqual(3L, message.Uid);
          Assert.IsTrue(message.IsSeen);
          Assert.IsFalse(message.IsRecent);
        }
      });
    }
#endif

    [Test]
    public void TestReadAs_Stream_WithConverter()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ret = message.ReadAs(delegate(Stream stream) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; while reading");

          return stream.ReadToEnd();
        });

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedOctets, ret.Length);
        Assert.AreEqual(expectedMessageBody,
                        OctetEncoding.EightBits.GetString(ret));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_ExceptionInConveter()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ex = Assert.Throws<NotSupportedException>(() => message.ReadAs((Converter<Stream, int>) delegate{
          throw new NotSupportedException("exception");
        }));

        Assert.AreEqual("exception", ex.Message);

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_ConverterNull()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        Assert.Throws<ArgumentNullException>(() => message.ReadAs<string>((Converter<Stream, string>)null));
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNull(server.DequeueRequest());
      });
    }

    [Test]
    public void TestReadAs_Stream_WithFunc1()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ret = message.ReadAs(delegate(Stream stream, string arg1) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);

          Assert.AreEqual("arg1", arg1);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; while reading");

          return stream.ReadToEnd();
        }, "arg1");

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedOctets, ret.Length);
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_WithFunc2()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ret = message.ReadAs(delegate(Stream stream, string arg1, string arg2) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);

          Assert.AreEqual("arg1", arg1);
          Assert.AreEqual("arg2", arg2);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; while reading");

          return stream.ReadToEnd();
        }, "arg1", "arg2");

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedOctets, ret.Length);
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_WithFunc3()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ret = message.ReadAs(delegate(Stream stream, string arg1, string arg2, string arg3) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);

          Assert.AreEqual("arg1", arg1);
          Assert.AreEqual("arg2", arg2);
          Assert.AreEqual("arg3", arg3);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; while reading");

          return stream.ReadToEnd();
        }, "arg1", "arg2", "arg3");

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedOctets, ret.Length);
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_WithFunc4()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ret = message.ReadAs(delegate(Stream stream, string arg1, string arg2, string arg3, string arg4) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(stream);
          Assert.AreEqual(expectedOctets, stream.Length);

          Assert.AreEqual("arg1", arg1);
          Assert.AreEqual("arg2", arg2);
          Assert.AreEqual("arg3", arg3);
          Assert.AreEqual("arg4", arg4);

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; while reading");

          return stream.ReadToEnd();
        }, "arg1", "arg2", "arg3", "arg4");

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedOctets, ret.Length);
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAllBytes()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        message.ReadAllBytes());

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSave()
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
          message.Save(outfile);

          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsTrue(File.Exists(outfile));
          Assert.AreEqual(expectedOctets, (new FileInfo(outfile)).Length);
          Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                          File.ReadAllBytes(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_OverwriteExistingFile()
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");
      const int existingFileLength = 10 * 1024;

      TestUtils.UsingFile(outfile, () => {
        File.WriteAllText(outfile, new string('x', existingFileLength));

        TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
          Assert.Less(expectedOctets, (new FileInfo(outfile)).Length, "less than existing file length");

          message.Save(outfile);

          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsTrue(File.Exists(outfile), "file existence");
          Assert.AreEqual(expectedOctets, (new FileInfo(outfile)).Length, "file length");
          Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                          File.ReadAllBytes(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_PathNull()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<ArgumentNullException>(() => message.Save(null));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSave_PathEmpty()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<ArgumentException>(() => message.Save(string.Empty));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestSave_FileNotCreatedOnException_MessageVanished()
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestCases.WithMessageVanished((server, message) => {
          Assert.Throws<ImapUnavailableException>(() => message.Save(outfile));

          Assert.IsFalse(File.Exists(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_FileNotCreatedOnException_MailboxClosed()
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestCases.WithMessageMailboxClosed((server, message) => {
          Assert.Throws<ImapUnavailableException>(() => message.Save(outfile));

          Assert.IsFalse(File.Exists(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test, Ignore("can't test")]
    public void TestSave_FileNotCreatedOnException_InvalidArgument()
    {
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");

      TestUtils.UsingFile(outfile, () => {
        TestCases.WithMessage((server, message) => {
          //Assert.Throws<ArgumentException>(() => message.Save(outfile, (IImapBodyStructure)null));

          Assert.IsFalse(File.Exists(outfile));

          Assert.IsFalse(message.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestWriteTo_Stream()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var outputStream = new MemoryStream();

        message.WriteTo(outputStream);

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.AreEqual(expectedOctets, outputStream.Length);

        outputStream.Dispose();

        Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        outputStream.ToArray());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteTo_Stream_DestinationNull()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<ArgumentNullException>(() => message.WriteTo((Stream)null));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteTo_Stream_DestinationNotWritable()
    {
      TestCases.WithMessage((server, message) => {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => message.WriteTo(stream));
          Assert.IsFalse(message.Client.IsBusy);
        }
      });
    }

    [Test]
    public void TestWriteTo_BinaryWriter()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var outputStream = new MemoryStream();
        var writer = new BinaryWriter(outputStream);

        message.WriteTo(writer);

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.AreEqual(expectedOctets, outputStream.Length);

        outputStream.Dispose();

        Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        outputStream.ToArray());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteTo_BinaryWriter_DestinationNull()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<ArgumentNullException>(() => message.WriteTo((BinaryWriter)null));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestDownload()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ret = message.Download();

        Assert.IsNotNull(ret);
        Assert.IsTrue(ret.CanSeek, "CanSeek");
        Assert.IsTrue(ret.CanRead, "CanRead");
        Assert.IsTrue(ret.CanWrite, "CanWrite");
        Assert.AreEqual(0L, ret.Position, "Position");
        Assert.AreEqual(expectedOctets, ret.Length, "Length");

        Assert.DoesNotThrow(() => ret.Seek(-1L, SeekOrigin.End));
        Assert.AreNotEqual(-1, ret.ReadByte());
        Assert.DoesNotThrow(() => ret.Seek(0L, SeekOrigin.Begin));

        Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        ret.ReadToEnd());

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var charset = Encoding.UTF8;

        using (var reader = message.OpenText(charset)) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(reader);
          Assert.AreSame(charset, reader.CurrentEncoding);
          Assert.AreEqual(expectedMessageBody, reader.ReadToEnd());

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenText_DefaultCharset()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        using (var reader = message.OpenText()) {
          StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

          Assert.IsNotNull(reader);
          Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
          Assert.AreEqual(expectedMessageBody, reader.ReadToEnd());

          Assert.IsTrue(message.Client.IsBusy, "IsBusy; still open");
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after close");
      });
    }

    [Test]
    public void TestOpenText_CharsetNull()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<ArgumentNullException>(() => message.OpenText((Encoding)null));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent()
    {
      TestCases.WithMessage((server, message) => {
        Assert.Throws<NotSupportedException>(() => message.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_StreamReader()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        Assert.AreEqual(expectedMessageBody, message.ReadAs(delegate(StreamReader reader) {
          Assert.IsTrue(message.Client.IsBusy, "IsBusy; while reading");

          return reader.ReadToEnd();
        }));

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_StreamReader_ExceptionInConveter()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var ex = Assert.Throws<NotSupportedException>(() => message.ReadAs((Converter<StreamReader, int>)delegate(StreamReader reader) {
          throw new NotSupportedException("exception");
        }));

        Assert.AreEqual("exception", ex.Message);

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_StreamReader_ConverterNull()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        Assert.Throws<ArgumentNullException>(() => message.ReadAs<string>((Converter<StreamReader, string>)null));
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNull(server.DequeueRequest());
      });
    }

    [Test]
    public void TestReadLines()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        var lines = message.ReadLines();

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsTrue(message.Client.IsBusy, "IsBusy; still enumerating");

        var reader = new StringReader(expectedMessageBody);

        foreach (var line in lines) {
          var expectedLine = reader.ReadLine();

          if (expectedLine == null)
            break;

          Assert.AreEqual(expectedLine, line);
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy; after enumeration");
      });
    }

    [Test]
    public void TestReadAllLines()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        Assert.AreEqual(expectedMessageBody.Replace("\r\n", "\n").Replace("\n", "\r\n").TrimEnd(),
                        string.Join("\r\n", message.ReadAllLines()));

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAllText()
    {
      TestFetchMessage((server, message, expectedMessageBody, expectedOctets) => {
        Assert.AreEqual(expectedMessageBody,
                        message.ReadAllText());

        StringAssert.EndsWith("UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.10240>)\r\n", server.DequeueRequest());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    private void TestFetchHeader(string messageHeader,
                                 string expectedFetchHeaderCommand,
                                 Action<ImapPseudoServer, ImapMessageInfo> action)
    {
      TestCases.WithOpenedMailbox("INBOX",
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS ())\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        server.DequeueRequest();

        var octets = OctetEncoding.EightBits.GetByteCount(messageHeader);

        server.EnqueueTaggedResponse(expectedFetchHeaderCommand,
                                     string.Format("* {0} FETCH (RFC822.SIZE {1} BODY[] {{{1}}}\r\n{2})\r\n",
                                                   message.Uid,
                                                   octets,
                                                   messageHeader) +
                                     "$tag OK done\r\n");

        action(server, message);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetRawHeader()
    {
      GetRawHeader(message => message.GetRawHeader());
    }

    [Test]
    public void TestGetRawHeaderAsync()
    {
      GetRawHeader(message => message.GetRawHeaderAsync().Result);
    }

    [Test]
    public void TestGetRawHeader_HeaderFieldsSpecified_Null()
    {
      GetRawHeader(message => message.GetRawHeader((IEnumerable<string>)null));
    }

    [Test]
    public void TestGetRawHeader_HeaderFieldsSpecified_Empty()
    {
      GetRawHeader(message => message.GetRawHeader(Array.Empty<string>()));
    }

    [Test]
    public void TestGetRawHeaderAsync_HeaderFieldsSpecified_Null()
    {
      GetRawHeader(message => message.GetRawHeaderAsync((IEnumerable<string>)null).Result);
    }

    [Test]
    public void TestGetRawHeaderAsync_HeaderFieldsSpecified_Empty()
    {
      GetRawHeader(message => message.GetRawHeaderAsync(Array.Empty<string>()).Result);
    }

    private void GetRawHeader(Func<ImapMessageInfo, string> testFunc)
    {
      const string messageHeader = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

";

      TestFetchHeader(messageHeader,
                      "UID FETCH 1 (BODY.PEEK[HEADER]<0.10240>)\r\n",
                      (server, message) => {
        Assert.AreEqual(messageHeader, testFunc(message));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetRawHeader_HeaderFieldsSpecified()
    {
      GetRawHeader_HeaderFieldsSpecified((message, headers) => message.GetRawHeader(headers));
    }

    [Test]
    public void TestGetRawHeaderAsync_HeaderFieldsSpecified()
    {
      GetRawHeader_HeaderFieldsSpecified((message, headers) => message.GetRawHeaderAsync(headers).Result);
    }

    private void GetRawHeader_HeaderFieldsSpecified(Func<ImapMessageInfo, string[], string> testFunc)
    {
      const string messageHeader = @"Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=

";

      TestFetchHeader(messageHeader,
                      "UID FETCH 1 (BODY.PEEK[HEADER.FIELDS (date from to subject)]<0.10240>)\r\n",
                      (server, message) => {
        Assert.AreEqual(messageHeader, testFunc(message, new[] {"date", "from", "to", "subject"}));

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetHeader()
    {
      GetHeader(message => message.GetHeader());
    }

    [Test]
    public void TestGetHeaderAsync()
    {
      GetHeader(message => message.GetHeaderAsync().Result);
    }

    [Test]
    public void TestGetHeader_HeaderFieldsSpecified_Null()
    {
      GetHeader(message => message.GetHeader((IEnumerable<string>)null));
    }

    [Test]
    public void TestGetHeader_HeaderFieldsSpecified_Empty()
    {
      GetHeader(message => message.GetHeader(Array.Empty<string>()));
    }

    [Test]
    public void TestGetHeaderAsync_HeaderFieldsSpecified_Null()
    {
      GetHeader(message => message.GetHeaderAsync((IEnumerable<string>)null).Result);
    }

    [Test]
    public void TestGetHeaderAsync_HeaderFieldsSpecified_Empty()
    {
      GetHeader(message => message.GetHeaderAsync(Array.Empty<string>()).Result);
    }

    private void GetHeader(Func<ImapMessageInfo, KeyValuePair<string, string>[]> testFunc)
    {
      const string messageHeader = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

";
      
      TestFetchHeader(messageHeader,
                      "UID FETCH 1 (BODY.PEEK[HEADER]<0.10240>)\r\n",
                      (server, message) => {
        var headers = testFunc(message);

        Assert.IsNotEmpty(headers);
        Assert.AreEqual(13, headers.Length);

        Assert.AreEqual("<test@example.net>", headers.Where(pair => pair.Key == "Return-Path").First().Value);
        Assert.AreEqual("1.0", headers.Where(pair => pair.Key == "MIME-Version").First().Value);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetHeader_HeaderFieldsSpecified()
    {
      GetHeader_HeaderFieldsSpecified((message, headers) => message.GetHeader(headers));
    }

    [Test]
    public void TestGetHeaderAsync_HeaderFieldsSpecified()
    {
      GetHeader_HeaderFieldsSpecified((message, headers) => message.GetHeaderAsync(headers).Result);
    }

    private void GetHeader_HeaderFieldsSpecified(Func<ImapMessageInfo, string[], KeyValuePair<string, string>[]> testFunc)
    {
      const string messageHeader = @"From: test <test@example.net>
To: hoge@example.net

";
      
      TestFetchHeader(messageHeader,
                      "UID FETCH 1 (BODY.PEEK[HEADER.FIELDS (From To X-NonExistent)]<0.10240>)\r\n",
                      (server, message) => {
        var headers = testFunc(message, new[] {"From", "To", "X-NonExistent"});

        Assert.IsNotEmpty(headers);
        Assert.AreEqual(2, headers.Length);

        Assert.IsTrue(headers.Where(pair => pair.Key == "From").Any());
        Assert.IsTrue(headers.Where(pair => pair.Key == "To").Any());
        Assert.IsFalse(headers.Where(pair => pair.Key == "X-NonExistent").Any());

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetHeader_Decode()
    {
      GetHeader_Decode(message => message.GetHeader(true));
    }

    [Test]
    public void TestGetHeaderAsync_Decode()
    {
      GetHeader_Decode(message => message.GetHeaderAsync(true).Result);
    }

    private void GetHeader_Decode(Func<ImapMessageInfo, KeyValuePair<string, string>[]> testFunc)
    {
      const string messageHeader = @"Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

";

      TestFetchHeader(messageHeader,
                      "UID FETCH 1 (BODY.PEEK[HEADER]<0.10240>)\r\n",
                      (server, message) => {
        var headers = testFunc(message);

        Assert.IsNotEmpty(headers);
        Assert.AreEqual(3, headers.Length);

        Assert.AreEqual("テスト", headers.Where(pair => pair.Key == "Subject").First().Value);
        Assert.AreEqual("text/plain; charset=ISO-2022-JP", headers.Where(pair => pair.Key == "Content-Type").First().Value);

        Assert.IsFalse(message.Client.IsBusy);
      });
    }
  }
}

