using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using NUnit.Framework;

using Smdn.IO;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoOperationsStoreTests {
    [Test]
    public void TestRefresh()
    {
      Refresh(message => message.Refresh());
    }

    [Test]
    public void TestRefreshAsync()
    {
      Refresh(message => message.RefreshAsync().Wait());
    }

    private void Refresh(Action<ImapMessageInfo> testAction)
    {
      var envelope = "(" +
                     "\"Tue, 23 Mar 2010 11:22:33 +0900\" " +
                     "\"subject\" " + 
                     "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                     "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                     "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                     "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                     "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                     "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                     "\"<in.reply.to@mail.example.com>\" " + 
                     "\"<message.id@mail.example.com>\" " + 
                     ")";
      var fetchResp = "* FETCH 1 (UID 1 BODYSTRUCTURE (\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL) " +
                      "ENVELOPE " + envelope + " " +
                      "INTERNALDATE \"23-Mar-2010 11:22:33 +0900\" RFC822.SIZE 1024 FLAGS (\\Seen))\r\n" +
                      "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp, ImapMessageFetchAttributeOptions.AllAttributes,
                             (server, message) => {
        Assert.IsTrue(message.IsSeen);
        Assert.IsFalse(message.IsMarkedAsDeleted);
        Assert.IsNotNull(message.Envelope);
        Assert.IsNotNull(message.MessagePart);

        var currentEnvelope = message.Envelope;
        var currentMessagePart = message.MessagePart;

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen \\Deleted))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(message);

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsTrue(message.IsSeen);
        Assert.IsTrue(message.IsMarkedAsDeleted);

        Assert.AreSame(currentEnvelope, message.Envelope);
        Assert.AreSame(currentMessagePart, message.MessagePart);
      });
    }

    [Test]
    public void TestRefresh_MailboxClosed()
    {
      Refresh_MailboxClosed(message => message.Refresh());
    }

    [Test]
    public void TestRefreshAsync_MailboxClosed()
    {
      Refresh_MailboxClosed(message => message.RefreshAsync().Wait());
    }

    private void Refresh_MailboxClosed(Action<ImapMessageInfo> testAction)
    {
      TestCases.WithMessageMailboxClosed((server, message) => {
        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapUnavailableException>(() => testAction(message));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestRefresh_VanishedMessage()
    {
      Refresh_VanishedMessage(message => message.Refresh());
    }

    [Test]
    public void TestRefreshAsync_VanishedMessage()
    {
      Refresh_VanishedMessage(message => message.RefreshAsync().Wait());
    }

    private void Refresh_VanishedMessage(Action<ImapMessageInfo> testAction)
    {
      TestCases.WithMessageVanished((server, message) => {
        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapUnavailableException>(() => testAction(message));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestRefresh_MessageVanishingByFetch()
    {
      Refresh_MessageVanishingByFetch(message => message.Refresh());
    }

    [Test]
    public void TestRefreshAsync_MessageVanishingByFetch()
    {
      Refresh_MessageVanishingByFetch(message => message.RefreshAsync().Wait());
    }

    private void Refresh_MessageVanishingByFetch(Action<ImapMessageInfo> testAction)
    {
      TestCases.WithMessage((server, message) => {
        Assert.AreNotEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsFalse(message.IsDeleted);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapMessageVanishedException>(() => testAction(message));

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(ex.VanishedMessage);
        Assert.AreSame(message, ex.VanishedMessage);

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapMessageVanishedException deserialized) {
          Assert.IsNull(deserialized.VanishedMessage);
        });

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsTrue(message.IsDeleted);
      });
    }

    [Test]
    public void TestRefresh_AttributesNotFetched()
    {
      TestCases.WithMessage((server, message) => {
        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
        });

        message.Refresh();

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(1, message.Flags.Count);
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(message.IsSeen);
      });
    }

    [Test]
    public void TestRefresh_DynamicAttributeNotFetched()
    {
      var envelope = "(" +
                     "\"Tue, 23 Mar 2010 11:22:33 +0900\" " +
                     "\"subject\" " + 
                     "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                     "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                     "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                     "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                     "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                     "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                     "\"<in.reply.to@mail.example.com>\" " + 
                     "\"<message.id@mail.example.com>\" " + 
                     ")";
      var fetchResp = "* FETCH 1 (UID 1 BODYSTRUCTURE (\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL) " +
                      "ENVELOPE " + envelope + " " +
                      "INTERNALDATE \"23-Mar-2010 11:22:33 +0900\" RFC822.SIZE 1024)\r\n" +
                      "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp, ImapMessageFetchAttributeOptions.StaticAttributes,
                             (server, message) => {
        Assert.IsNotNull(message.Envelope);
        Assert.IsNotNull(message.MessagePart);

        var currentEnvelope = message.Envelope;
        var currentMessagePart = message.MessagePart;

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen \\Deleted))\r\n" +
                                       "$tag OK done\r\n");
        });

        message.Refresh();

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(2, message.Flags.Count);
        Assert.IsTrue(message.IsSeen);
        Assert.IsTrue(message.IsMarkedAsDeleted);

        Assert.AreSame(currentEnvelope, message.Envelope);
        Assert.AreSame(currentMessagePart, message.MessagePart);
      });
    }

    [Test]
    public void TestRefresh_CondStoreCapable()
    {
      var selectResp =
        "* OK [HIGHESTMODSEQ 1]\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.CondStore},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        Assert.IsTrue(mailbox.IsModSequencesAvailable);

        server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS MODSEQ)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Seen) MODSEQ (1))\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.IsTrue(message.IsSeen);
        Assert.IsFalse(message.IsMarkedAsDeleted);
        Assert.AreEqual(1L, message.ModSeq);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS MODSEQ)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen \\Deleted) MODSEQ (2))\r\n" +
                                       "$tag OK done\r\n");
        });

        message.Refresh();

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsTrue(message.IsSeen);
        Assert.IsTrue(message.IsMarkedAsDeleted);
        Assert.AreEqual(2L, message.ModSeq);
      });
    }

    [Test]
    public void TestRefresh_CondStoreCapable_NoModSeq()
    {
      var selectResp =
        "* OK [NOMODSEQ]\r\n" +
        "* OK [READ-WRITE]\r\n" +
        "$tag OK done\r\n";

      TestCases.WithOpenedMailbox(new[] {ImapCapability.CondStore},
                                   "INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        Assert.IsFalse(mailbox.IsModSequencesAvailable);

        server.EnqueueTaggedResponse("UID FETCH 1 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        var message = mailbox.GetMessageByUid(1L, ImapMessageFetchAttributeOptions.DynamicAttributes);

        Assert.IsTrue(message.IsSeen);
        Assert.IsFalse(message.IsMarkedAsDeleted);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen \\Deleted))\r\n" +
                                       "$tag OK done\r\n");
        });

        message.Refresh();

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsTrue(message.IsSeen);
        Assert.IsTrue(message.IsMarkedAsDeleted);
      });
    }

    [Test, Ignore("this will never happen")]
    public void TestRefreshSequenceChanged()
    {
      var fetchResp =
        "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
        "* FETCH 2 (UID 2 FLAGS (\\Seen))\r\n" +
        "* FETCH 3 (UID 3 FLAGS (\\Recent))\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessages(fetchResp,
                             ImapMessageFetchAttributeOptions.DynamicAttributes,
                             (server, messages) => {
        Assert.AreEqual(1, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Recent));
        Assert.IsTrue(messages[2].IsRecent);

        // FETCH
        server.EnqueueTaggedResponse("* EXPUNGE 1\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     "* FETCH 1 (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        messages[2].Refresh();

        StringAssert.EndsWith("UID FETCH 3 (FLAGS)\r\n", server.DequeueRequest());

        Assert.IsTrue (messages[0].IsDeleted);
        Assert.IsTrue (messages[1].IsDeleted);
        Assert.IsFalse(messages[2].IsDeleted);

        var message = messages[2];

        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(1, message.Flags.Count);
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(message.IsSeen);
      });
    }

    [Test]
    public void TestEnsureDynamicAttributesFetched()
    {
      TestCases.WithMessage((server, message) => {
        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "* FETCH 1 (UID 1 FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.IsTrue(message.IsSeen);

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(1, message.Flags.Count);
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(message.IsSeen);
      });
    }

    [Test]
    public void TestEnsureDynamicAttributesFetched_MailboxClosed()
    {
      TestCases.WithMessageMailboxClosed((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => Assert.IsTrue(message.IsSeen));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestEnsureDynamicAttributesFetched_VanishedMessage()
    {
      TestCases.WithMessageVanished((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => Assert.IsTrue(message.IsSeen));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestEnsureDynamicAttributesFetched_MessageVanishingByFetch()
    {
      TestCases.WithMessage((server, message) => {
        Assert.AreNotEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsFalse(message.IsDeleted);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMessageVanishedException>(() => Assert.IsTrue(message.IsSeen));

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(ex.VanishedMessage);
        Assert.AreSame(message, ex.VanishedMessage);

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapMessageVanishedException deserialized) {
          Assert.IsNull(deserialized.VanishedMessage);
        });

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsTrue(message.IsDeleted);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));
        Assert.IsFalse(message.IsAvailable);
      });
    }

#if false
    [Test, Ignore("this will never happen")]
    public void TestEnsureDynamicAttributesFetchedSequenceChanged()
    {
      var fetchResp =
        "* FETCH 1 (UID 1)\r\n" +
        "* FETCH 2 (UID 2)\r\n" +
        "* FETCH 3 (UID 3)\r\n" +
        "$tag OK done\r\n";

      TestMessages(fetchResp, ImapMessageFetchAttributeOptions.None,
                   delegate(ImapPseudoServer server, ImapMessageInfo[] messages) {
        // FETCH
        server.EnqueueTaggedResponse("* EXPUNGE 1\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     "* FETCH 1 (UID 3 FLAGS (\\Seen))\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(messages[2].IsSeen);

        StringAssert.EndsWith("UID FETCH 3 (FLAGS)\r\n", server.DequeueRequest());

        Assert.IsTrue (messages[0].IsDeleted);
        Assert.IsTrue (messages[1].IsDeleted);
        Assert.IsFalse(messages[2].IsDeleted);

        var message = messages[2];

        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.IsNotNull(message.Flags);
        Assert.AreEqual(1, message.Flags.Count);
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(message.IsSeen);
      });
    }
#endif

    [Test]
    public void TestEnsureStaticAttributesFetched()
    {
      TestCases.WithMessage((server, message) => {
        var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
        var envelope = "(" +
                       "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
                       "\"test message\" " + 
                       "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                       "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                       "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                       "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                       "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                       "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                       "\"<in.reply.to@mail.example.com>\" " + 
                       "\"<message.id@mail.example.com>\" " + 
                       ")";
        var internaldate = "\"23-Mar-2010 09:46:52 +0900\"";

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)\r\n",
                                       "* FETCH 1 (" +
                                       "UID 1 " +
                                       "BODYSTRUCTURE " + bodystructure +
                                       "ENVELOPE " + envelope +
                                       "INTERNALDATE " + internaldate +
                                       "RFC822.SIZE 1024 " +
                                       ")\r\n" +
                                       "$tag OK done\r\n");
        });

        Assert.AreEqual(1024L, message.Length);

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.MessagePart);
        Assert.IsNotNull(message.Envelope);
        Assert.AreNotEqual(default(DateTimeOffset), message.DeliveredDate);
      });
    }

    [Test]
    public void TestEnsureStaticAttributesFetched_MailboxClosed()
    {
      TestCases.WithMessageMailboxClosed((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => Assert.AreSame(0, message.Length));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestEnsureStaticAttributesFetched_VanishedMessage()
    {
      TestCases.WithMessageVanished((server, message) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => Assert.AreEqual(0, message.Length));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestEnsureStaticAttributesFetched_MessageVanishingByFetch()
    {
      TestCases.WithMessage((server, message) => {
        Assert.AreNotEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsFalse(message.IsDeleted);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)\r\n",
                                       "$tag OK done\r\n");
        });

        var ex = Assert.Throws<ImapMessageVanishedException>(() => Assert.AreEqual(0, message.Length));

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(ex.VanishedMessage);
        Assert.AreSame(message, ex.VanishedMessage);

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapMessageVanishedException deserialized) {
          Assert.IsNull(deserialized.VanishedMessage);
        });

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, message.Sequence);
        Assert.IsTrue(message.IsDeleted);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(message.Mailbox));
        Assert.IsFalse(message.IsAvailable);
      });
    }

#if false
    [Test, Ignore("this will never happen")]
    public void TestEnsureStaticAttributesFetchedSequenceChanged()
    {
      var fetchResp =
        "* FETCH 1 (UID 1)\r\n" +
        "* FETCH 2 (UID 2)\r\n" +
        "* FETCH 3 (UID 3)\r\n" +
        "$tag OK done\r\n";

      TestMessages(fetchResp, ImapMessageFetchAttributeOptions.None,
                   delegate(ImapPseudoServer server, ImapMessageInfo[] messages) {
        var bodystructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2109 54 NIL NIL NIL NIL)";
        var envelope = "(" +
                       "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
                       "\"test message\" " + 
                       "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
                       "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
                       "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
                       "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
                       "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
                       "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
                       "\"<in.reply.to@mail.example.com>\" " + 
                       "\"<message.id@mail.example.com>\" " + 
                       ")";
        var internaldate = "\"23-Mar-2010 09:46:52 +0900\"";

        // FETCH
        server.EnqueueTaggedResponse("* EXPUNGE 1\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "* EXISTS 1\r\n" +
                                     "* FETCH 1 (" +
                                     "UID 3 " +
                                     "BODYSTRUCTURE " + bodystructure +
                                     "ENVELOPE " + envelope +
                                     "INTERNALDATE " + internaldate +
                                     "RFC822.SIZE 1024 " +
                                     ")\r\n" +
                                     "$tag OK done\r\n");

        Assert.AreEqual(1024L, messages[2].Length);

        StringAssert.EndsWith("UID FETCH 3 (BODYSTRUCTURE ENVELOPE INTERNALDATE RFC822.SIZE)\r\n", server.DequeueRequest());

        Assert.IsTrue (messages[0].IsDeleted);
        Assert.IsTrue (messages[1].IsDeleted);
        Assert.IsFalse(messages[2].IsDeleted);

        var message = messages[2];

        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.IsNotNull(message.BodyStructure);
        Assert.IsNotNull(message.Envelope);
        Assert.AreEqual(new DateTimeOffset(2010, 03, 23, 09, 46, 52, TimeSpan.FromHours(+9)),
                        message.InternalDate);
      });
    }
#endif

    [Test]
    public void TestStore_VanishedMessage()
    {
      Store_VanishedMessage((message, item) => message.Store(item));
    }

    [Test]
    public void TestStoreAsync_VanishedMessage()
    {
      Store_VanishedMessage((message, item) => message.StoreAsync(item).Wait());
    }

    private void Store_VanishedMessage(Action<ImapMessageInfo, ImapStoreDataItem> testAction)
    {
      TestCases.WithMessageVanished((server, message) => {
        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapUnavailableException>(() => testAction(message, ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen)));

        Assert.IsFalse(message.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestToggleFlags()
    {
      ToggleFlags((message, f1, f2) => message.ToggleFlags(f1, f2));
    }

    [Test]
    public void TestToggleFlagsAsync()
    {
      ToggleFlags((message, f1, f2) => message.ToggleFlagsAsync(f1, f2).Wait());
    }

    private void ToggleFlags(Action<ImapMessageInfo, ImapMessageFlag, ImapMessageFlag> testAction)
    {
      var fetchResp =
        "* FETCH 1 (UID 1 FLAGS (\\Seen $label1))\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp, ImapMessageFetchAttributeOptions.DynamicAttributes,
                             (server, message) => {
        Assert.AreEqual(2, message.Flags.Count);
        Assert.IsTrue(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(message.Flags.Contains("$label1"));

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS (\\Seen)\r\n",
                                       "* 1 FETCH (FLAGS ($label1))\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Draft)\r\n",
                                       "* 1 FETCH (FLAGS (\\Draft $label1))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(message, ImapMessageFlag.Seen, ImapMessageFlag.Draft);

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.AreEqual(2, message.Flags.Count);
        Assert.IsFalse(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue (message.Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue (message.Flags.Contains("$label1"));
        Assert.IsFalse(message.IsSeen);

        responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS ($label1)\r\n",
                                       "* 1 FETCH (FLAGS (\\Draft))\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS ($label2 $label3)\r\n",
                                       "* 1 FETCH (FLAGS (\\Draft $label2 $label3))\r\n" +
                                       "$tag OK done\r\n");
        });

        message.ToggleKeywords("$label1", "$label2", "$label3");

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.AreEqual(3, message.Flags.Count);
        Assert.IsFalse(message.Flags.Contains("$label1"));
        Assert.IsTrue (message.Flags.Contains("$label2"));
        Assert.IsTrue (message.Flags.Contains("$label3"));
        Assert.IsTrue (message.Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(message.IsSeen);
      });
    }

    [Test]
    public void TestToggleFlags_DynamicAttributeNotFetched()
    {
      ToggleFlags_DynamicAttributeNotFetched((message, flag) => message.ToggleFlags(flag));
    }

    [Test]
    public void TestToggleFlagsAsync_DynamicAttributeNotFetched()
    {
      ToggleFlags_DynamicAttributeNotFetched((message, flag) => message.ToggleFlagsAsync(flag).Wait());
    }

    private void ToggleFlags_DynamicAttributeNotFetched(Action<ImapMessageInfo, ImapMessageFlag> testAction)
    {
      TestCases.WithMessage((server, message) => {
        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (FLAGS)\r\n",
                                       "* 1 FETCH (FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS (\\Seen)\r\n",
                                       "* 1 FETCH (FLAGS ())\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(message, ImapMessageFlag.Seen);

        responseTask.Wait();

        Assert.IsFalse(message.Client.IsBusy);

        Assert.AreEqual(0, message.Flags.Count);
        Assert.IsFalse(message.Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsFalse(message.IsSeen);
      });
    }
  }
}

