using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoTests {
    [Test]
    public void TestRefresh()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* RECENT 1\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);
        Assert.AreEqual(1L, mailbox.RecentMessageCount);
        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(5, mailbox.ApplicableFlags.Count);
        Assert.IsFalse(mailbox.ApplicableFlags.Contains("$label1"));

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* EXPUNGE 1\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "* EXISTS 1\r\n" +
                                       "* RECENT 2\r\n" +
                                       "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft $label1)\r\n" +
                                       "$tag OK done\r\n");
        });

        mailbox.Refresh();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(1L, mailbox.ExistMessageCount);
        Assert.AreEqual(2L, mailbox.RecentMessageCount);
        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(6, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("$label1"));
      });
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags()
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* RECENT 1\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        CollectionAssert.DoesNotContain(mailbox.Flags, ImapMailboxFlag.Marked);

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX",
                                       "* LIST (\\Marked) \"/\" \"INBOX\"\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.Refresh(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Marked);
      });
    }

    [Test]
    public void TestRefresh_Closed()
    {
      TestCases.WithClosedMailbox("INBOX",
                                   (server, mailbox) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => mailbox.Refresh());

        Assert.IsFalse(mailbox.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestOpen()
    {
      Open(false, mailbox => mailbox.Open());
    }

    [Test]
    public void TestOpenAsync()
    {
      Open(false, mailbox => mailbox.OpenAsync().Result);
    }

    [Test]
    public void TestOpen_AsReadOnly()
    {
      Open(true, mailbox => mailbox.Open(true));
    }

    [Test]
    public void TestOpenAsync_AsReadOnly()
    {
      Open(true, mailbox => mailbox.OpenAsync(true).Result);
    }

    private void Open(bool readOnly, Func<ImapOpenedMailboxInfo, ImapOpenedMailboxInfo> testFunc)
    {
      var selectResp = readOnly
        ? "$tag OK [READ-ONLY] done\r\n"
        : "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX",
                                   selectResp,
                                   (server, mailbox) => {
        Assert.AreEqual(readOnly, mailbox.IsReadOnly);

        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, testFunc(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox, mailbox.Client.OpeningMailbox);

        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreEqual(readOnly, mailbox.IsReadOnly);
      });
    }

    [Test]
    public void TestOpen_UidNotSticky()
    {
      Open_UidNotSticky(mailbox => mailbox.Open());
    }

    [Test]
    public void TestOpenAsync_UidNotSticky()
    {
      Open_UidNotSticky(mailbox => mailbox.OpenAsync().Result);
    }

    private void Open_UidNotSticky(Func<ImapOpenedMailboxInfo, ImapOpenedMailboxInfo> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "* NO [UIDNOTSTICKY]\r\n" +
        "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
        "* OK [PERMANENTFLAGS (\\Deleted \\Seen)] Limited\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.IsTrue(mailbox.IsOpen);

        Assert.AreSame(mailbox, testFunc(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox, mailbox.Client.OpeningMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsFalse(mailbox.IsReadOnly);
        Assert.AreEqual(23L, mailbox.UidValidity);
        Assert.IsFalse(mailbox.IsUidPersistent);

        Assert.IsNotNull(mailbox.ApplicableFlags);
        Assert.AreEqual(5, mailbox.ApplicableFlags.Count);

        Assert.IsNotNull(mailbox.PermanentFlags);
        Assert.AreEqual(2, mailbox.PermanentFlags.Count);
        Assert.IsFalse(mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsFalse(mailbox.IsAllowedToCreateKeywords);
      });
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_OpenNewly()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_ReopenClosed()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_ReopenReadWriteChange()
    {
    }

    [Test]
    public void TestOpen_ReopenMailbox()
    {
      TestCases.WithOpenedMailbox("INBOX",
                                   (server, mailbox) => {
        Assert.IsTrue(mailbox.IsOpen);
        Assert.AreSame(mailbox, mailbox.Client.OpeningMailbox);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        mailbox.Close();

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       string.Format("* LIST () \"\" {0}\r\n", mailbox.FullName) +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                       "$tag OK [READ-WRITE] done\r\n"); 
        });

        var reopenedMailbox = mailbox.Client.OpenMailbox(mailbox.FullName);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(mailbox.Client.OpeningMailbox, reopenedMailbox);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.AreNotSame(mailbox, mailbox.Client.OpeningMailbox);
      });
    }

    [Test]
    public void TestClose()
    {
      Close(mailbox => mailbox.Close());
    }

    [Test]
    public void TestCloseAsync()
    {
      Close(mailbox => mailbox.CloseAsync().Wait());
    }

    private void Close(Action<ImapOpenedMailboxInfo> testAction)
    {
      TestCases.WithOpenedMailbox("INBOX",
                                   (server, mailbox) => {
        Assert.IsTrue(mailbox.IsOpen);

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpeningMailbox);
      });
    }

    [Test]
    public void TestClose_AlreadyClosed()
    {
      Close_AlreadyClosed(mailbox => mailbox.Close());
    }

    [Test]
    public void TestCloseAsync_AlreadyClosed()
    {
      Close_AlreadyClosed(mailbox => mailbox.CloseAsync().Wait());
    }

    private void Close_AlreadyClosed(Action<ImapOpenedMailboxInfo> testAction)
    {
      TestCases.WithClosedMailbox("INBOX",
                                   (server, mailbox) => {
        Assert.IsFalse(mailbox.IsOpen);

        Assert.DoesNotThrow(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(mailbox.Client.OpeningMailbox);
      });
    }

    [Test]
    public void TestClose_SameMailboxOpened()
    {
      CloseDisposeSameMailboxOpened(true);
    }

    [Test]
    public void TestDispose()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var mailbox = client.OpenInbox();

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        });

        Assert.DoesNotThrow(() => (mailbox as IDisposable).Dispose());

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(client.OpeningMailbox);
      });
    }

    [Test]
    public void TestDispose_AlreadyClosed()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var mailbox = client.OpenInbox();

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        client.CloseMailbox();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(client.OpeningMailbox);

        Assert.DoesNotThrow(() => (mailbox as IDisposable).Dispose());

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsNull(client.OpeningMailbox);
      });
    }

    [Test]
    public void TestDispose_AlreadyLoggedOut()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var mailbox = client.OpenInbox();

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK closed\r\n");
        server.EnqueueUntaggedByeResponse("* BYE logged out\r\n"); // LOGOUT

        try {
          client.Logout();
        }
        catch (Smdn.Net.Imap4.Protocol.ImapConnectionException) {
          // ignore disconnected
        }

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(client.IsConnected);

        Assert.DoesNotThrow(() => (mailbox as IDisposable).Dispose());
      });
    }

    [Test]
    public void TestDispose_SameMailboxOpened()
    {
      CloseDisposeSameMailboxOpened(false);
    }

    private void CloseDisposeSameMailboxOpened(bool close)
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var firstOpenedMailbox = client.OpenInbox();

        Assert.AreSame(client.OpeningMailbox, firstOpenedMailbox);
        Assert.IsTrue(firstOpenedMailbox.IsOpen);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        firstOpenedMailbox.Close();

        Assert.IsNull(client.OpeningMailbox);
        Assert.IsFalse(firstOpenedMailbox.IsOpen);

        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK [READ-WRITE] done\r\n");

        var secondOpenedMailbox = client.OpenInbox();

        Assert.AreNotSame(client.OpeningMailbox, firstOpenedMailbox);
        Assert.AreSame(client.OpeningMailbox, secondOpenedMailbox);
        Assert.IsFalse(firstOpenedMailbox.IsOpen);
        Assert.IsTrue(secondOpenedMailbox.IsOpen);

        if (close)
          Assert.DoesNotThrow(() => firstOpenedMailbox.Close());
        else
          Assert.DoesNotThrow(() => (firstOpenedMailbox as IDisposable).Dispose());

        Assert.IsFalse(firstOpenedMailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestExpunge()
    {
      Expunge(mailbox => mailbox.Expunge());
    }

    [Test]
    public void TestExpungeAsync()
    {
      Expunge(mailbox => mailbox.ExpungeAsync().Wait());
    }

    private void Expunge(Action<ImapOpenedMailboxInfo> testAction)
    {
      var selectResp =
        "* EXISTS 3\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.AreEqual(3L, mailbox.ExistMessageCount);

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID FLAGS)\r\n",
                                     "* FETCH 1 (UID 1 FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (UID 2 FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 3 (UID 3 FLAGS (\\Deleted))\r\n" +
                                     "$tag OK done\r\n");

        var messages = mailbox.GetMessages(ImapMessageFetchAttributeOptions.DynamicAttributes).ToArray();

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);
        Assert.IsFalse(messages[0].IsDeleted);

        Assert.AreEqual(2L, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsFalse(messages[1].IsDeleted);

        Assert.AreEqual(3L, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        Assert.IsFalse(messages[2].IsDeleted);

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("EXPUNGE\r\n",
                                       "* EXPUNGE 3\r\n" +
                                       "* EXPUNGE 2\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "* EXISTS 0\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0L, mailbox.ExistMessageCount);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.AreEqual(1L, messages[0].Uid);
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.AreEqual(2L, messages[1].Uid);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.AreEqual(3L, messages[2].Uid);
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestExpunge_Closed()
    {
      Expunge_Closed(mailbox => mailbox.Expunge());
    }

    [Test]
    public void TestExpungeAsync_Closed()
    {
      Expunge_Closed(mailbox => mailbox.ExpungeAsync().Wait());
    }

    private void Expunge_Closed(Action<ImapOpenedMailboxInfo> testAction)
    {
      TestCases.WithClosedMailbox("INBOX",
                                   (server, mailbox) => {
        var ex = Assert.Throws<ImapUnavailableException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestMoveMessagesTo()
    {
      CopyOrMoveMessagesTo(true,
                           (mailbox, dest) => mailbox.MoveMessagesTo(dest));
    }

    [Test]
    public void TestMoveMessagesToAsync()
    {
      CopyOrMoveMessagesTo(true,
                           (mailbox, dest) => mailbox.MoveMessagesToAsync(dest).Wait());
    }

    [Test]
    public void TestCopyMessagesTo()
    {
      CopyOrMoveMessagesTo(false,
                           (mailbox, dest) => mailbox.CopyMessagesTo(dest));
    }

    [Test]
    public void TestCopyMessagesToAsync()
    {
      CopyOrMoveMessagesTo(false,
                           (mailbox, dest) => mailbox.CopyMessagesToAsync(dest).Wait());
    }

    private void CopyOrMoveMessagesTo(bool move,
                                      Action<ImapOpenedMailboxInfo, ImapMailboxInfo> testAction)
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = mailbox.Client.GetMailbox("dest");

        var responseTask = TestCases.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("COPY 1:5 dest\r\n",
                                       "$tag OK done\r\n");

          if (move)
            server.EnqueueTaggedResponse("STORE 1:5 +FLAGS (\\Deleted)\r\n",
                                         "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 2 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 3 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 4 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 5 (FLAGS (\\Deleted))\r\n" +
                                         "$tag OK done\r\n");
        });

        testAction(mailbox, destMailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveMessagesTo_DestinationNull()
    {
      CopyOrMoveMessagesTo_DestinationNull((mailbox, dest) => mailbox.MoveMessagesTo(dest));
    }

    [Test]
    public void TestMoveMessagesToAsync_DestinationNull()
    {
      CopyOrMoveMessagesTo_DestinationNull((mailbox, dest) => mailbox.MoveMessagesToAsync(dest));
    }

    [Test]
    public void TestCopyMessagesTo_DestinationNull()
    {
      CopyOrMoveMessagesTo_DestinationNull((mailbox, dest) => mailbox.CopyMessagesTo(dest));
    }

    [Test]
    public void TestCopyMessagesToAsync_DestinationNull()
    {
      CopyOrMoveMessagesTo_DestinationNull((mailbox, dest) => mailbox.CopyMessagesToAsync(dest));
    }

    private void CopyOrMoveMessagesTo_DestinationNull(Action<ImapOpenedMailboxInfo, ImapMailboxInfo> testAction)
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp,
                                   (server, mailbox) => {
        Assert.Throws<ArgumentNullException>(() => testAction(mailbox, null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveMessagesTo_MoveToMailboxOfDifferentSession()
    {
      CopyOrMoveMessagesToMailboxOfDifferentSession(true,
                                                    (mailbox, dest) => mailbox.MoveMessagesTo(dest));
    }

    [Test]
    public void TestMoveMessagesToAsync_MoveToMailboxOfDifferentSession()
    {
      CopyOrMoveMessagesToMailboxOfDifferentSession(true,
                                                    (mailbox, dest) => mailbox.MoveMessagesToAsync(dest).Wait());
    }

    [Test]
    public void TestCopyMessagesTo_CopyToMailboxOfDifferentSession()
    {
      CopyOrMoveMessagesToMailboxOfDifferentSession(false,
                                                    (mailbox, dest) => mailbox.CopyMessagesTo(dest));
    }

    [Test]
    public void TestCopyMessagesToAsync_CopyToMailboxOfDifferentSession()
    {
      CopyOrMoveMessagesToMailboxOfDifferentSession(false,
                                                    (mailbox, dest) => mailbox.CopyMessagesToAsync(dest).Wait());
    }

    private void CopyOrMoveMessagesToMailboxOfDifferentSession(bool move,
                                                               Action<ImapOpenedMailboxInfo, ImapMailboxInfo> testAction)
    {
      var selectResp =
        "* 3 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestCases.WithOpenedMailbox("INBOX", selectResp, (serverSource, mailboxSource) => {
        TestCases.WithAuthenticatedClient((serverDest, clientDest) => {
          serverDest.EnqueueTaggedResponse("LIST ",
                                           "* LIST () \"\" dest\r\n" +
                                           "$tag OK done\r\n");

          var mailboxDest = clientDest.GetMailbox("dest");

          serverDest.DequeueRequest(); // LIST

          Assert.AreNotSame(mailboxSource.Client, mailboxDest.Client);

          var responseTask = TestCases.QueueResponseTask(serverSource, mailboxSource.Client, delegate {
            // NOOP (source)
            serverSource.EnqueueTaggedResponse("NOOP\r\n",
                                               "$tag OK done\r\n");
            // FETCH (source)
            serverSource.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                               "* FETCH 1 (UID 1)\r\n" +
                                               "* FETCH 2 (UID 2)\r\n" +
                                               "* FETCH 3 (UID 3)\r\n" +
                                               "$tag OK done\r\n");
            serverSource.EnqueueTaggedResponse("UID FETCH 1 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                               "* FETCH 1 (" +
                                               "FLAGS (\\Answered $label1) " +
                                               "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                               "RFC822.SIZE 13 " +
                                               "BODY[] {13}\r\ntest message1)\r\n" +
                                               "$tag OK done\r\n");
            serverSource.EnqueueTaggedResponse("UID FETCH 2 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                               "* FETCH 2 (" +
                                               "FLAGS (\\Answered $label2) " +
                                               "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                               "RFC822.SIZE 13 " +
                                               "BODY[] {13}\r\ntest message2)\r\n" +
                                               "$tag OK done\r\n");
            serverSource.EnqueueTaggedResponse("UID FETCH 3 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                               "* FETCH 3 (" +
                                               "FLAGS (\\Answered $label3) " +
                                               "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                               "RFC822.SIZE 13 " +
                                               "BODY[] {13}\r\ntest message3)\r\n" +
                                               "$tag OK done\r\n");

            if (move)
              // UID STORE (source)
              serverSource.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                                 "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                                 "* FETCH 2 (FLAGS (\\Deleted))\r\n" +
                                                 "* FETCH 3 (FLAGS (\\Deleted))\r\n" +
                                                 "$tag OK done\r\n");
          });

          // APPEND (dest)
          serverDest.EnqueueTaggedResponse("APPEND dest (\\Answered $label1) \"25-Jan-2011 15:29:06 +0900\" {13}\r\n" +
                                           "test message1\r\n",
                                           "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
          serverDest.EnqueueTaggedResponse("APPEND dest (\\Answered $label2) \"25-Jan-2011 15:29:06 +0900\" {13}\r\n" +
                                           "test message2\r\n",
                                           "$tag OK [APPENDUID 38505 3956] APPEND completed\r\n");
          serverDest.EnqueueTaggedResponse("APPEND dest (\\Answered $label3) \"25-Jan-2011 15:29:06 +0900\" {13}\r\n" +
                                           "test message3\r\n",
                                           "$tag OK [APPENDUID 38505 3957] APPEND completed\r\n");

          testAction(mailboxSource, mailboxDest);

          responseTask.Wait();

          Assert.IsFalse(mailboxSource.Client.IsBusy);
        });
      });
    }
  }
}
